import atexit
import operator
import os
import sys
import pickle
import shutil
import tempfile
import textwrap
import unittest

import numpy as np
import pytest

try:
    import ipytree
except ImportError:  # pragma: no cover
    ipytree = None

from numcodecs import Zlib
from numpy.testing import assert_array_equal

from zarr._storage.store import _get_metadata_suffix, v3_api_available
from zarr.attrs import Attributes
from zarr.core import Array
from zarr.creation import open_array
from zarr.hierarchy import Group, group, open_group
from zarr.storage import (
    ABSStore,
    DBMStore,
    KVStore,
    DirectoryStore,
    FSStore,
    LMDBStore,
    LRUStoreCache,
    MemoryStore,
    NestedDirectoryStore,
    SQLiteStore,
    ZipStore,
    array_meta_key,
    atexit_rmglob,
    atexit_rmtree,
    data_root,
    group_meta_key,
    init_array,
    init_group,
    meta_root,
)
from zarr._storage.v3 import (
    ABSStoreV3,
    KVStoreV3,
    DirectoryStoreV3,
    MemoryStoreV3,
    FSStoreV3,
    ZipStoreV3,
    DBMStoreV3,
    LMDBStoreV3,
    SQLiteStoreV3,
    LRUStoreCacheV3,
)
from zarr.util import InfoReporter, buffer_size
from zarr.tests.util import skip_test_env_var, have_fsspec, abs_container, mktemp


_VERSIONS = (2, 3) if v3_api_available else (2,)

# noinspection PyStatementEffect


class TestGroup(unittest.TestCase):
    @staticmethod
    def create_store():
        # can be overridden in sub-classes
        return KVStore(dict()), None

    def create_group(
        self, store=None, path=None, read_only=False, chunk_store=None, synchronizer=None
    ):
        # can be overridden in sub-classes
        if store is None:
            store, chunk_store = self.create_store()
        init_group(store, path=path, chunk_store=chunk_store)
        g = Group(
            store,
            path=path,
            read_only=read_only,
            chunk_store=chunk_store,
            synchronizer=synchronizer,
        )
        return g

    def test_ipython_repr_methods(self):
        g = self.create_group()
        for method in [
            "html",
            "json",
            "javascript",
            "markdown",
            "svg",
            "png",
            "jpeg",
            "latex",
            "pdf",
            "mimebundle",
        ]:
            assert operator.methodcaller(f"_repr_{method}_")(g) is None
        with pytest.raises(AttributeError):
            g._ipython_display_()
        with pytest.raises(AttributeError):
            g._ipython_canary_method_should_not_exist_()

    def test_group_init_1(self):
        store, chunk_store = self.create_store()
        g = self.create_group(store, chunk_store=chunk_store)
        assert store is g.store
        if chunk_store is None:
            assert store is g.chunk_store
        else:
            assert chunk_store is g.chunk_store
        assert not g.read_only
        assert "" == g.path
        assert "/" == g.name
        assert "" == g.basename
        assert isinstance(g.attrs, Attributes)
        g.attrs["foo"] = "bar"
        assert g.attrs["foo"] == "bar"
        assert isinstance(g.info, InfoReporter)
        assert isinstance(repr(g.info), str)
        assert isinstance(g.info._repr_html_(), str)
        store.close()

    def test_group_init_2(self):
        store, chunk_store = self.create_store()
        g = self.create_group(store, chunk_store=chunk_store, path="/foo/bar/", read_only=True)
        assert store is g.store
        assert g.read_only
        assert "foo/bar" == g.path
        assert "/foo/bar" == g.name
        assert "bar" == g.basename
        assert isinstance(g.attrs, Attributes)
        store.close()

    def test_group_init_errors_1(self):
        store, chunk_store = self.create_store()
        # group metadata not initialized
        with pytest.raises(ValueError):
            Group(store, chunk_store=chunk_store)
        store.close()

    def test_group_repr(self):
        store, chunk_store = self.create_store()
        g = self.create_group(store, chunk_store=chunk_store)
        assert g.name in repr(g)

    def test_group_init_errors_2(self):
        store, chunk_store = self.create_store()
        init_array(store, shape=1000, chunks=100, chunk_store=chunk_store)
        # array blocks group
        with pytest.raises(ValueError):
            Group(store, chunk_store=chunk_store)
        store.close()

    def _subgroup_path(self, group, path):
        path = path.rstrip("/")
        group_path = "/".join([group.path, path])
        group_path = group_path.lstrip("/")
        group_name = "/" + group_path
        return group_path, group_name

    def test_create_group(self):
        g1 = self.create_group()

        if g1._version == 2:
            path, name = "", "/"
        else:
            path, name = "group", "/group"
        # check root group
        assert path == g1.path
        assert name == g1.name

        # create level 1 child group
        g2 = g1.create_group("foo")
        # check with relative path
        path, name = self._subgroup_path(g1, "foo")
        assert isinstance(g2, Group)
        assert path == g2.path
        assert name == g2.name

        # create level 2 child group
        g3 = g2.create_group("bar")
        path, name = self._subgroup_path(g2, "bar")
        assert isinstance(g3, Group)
        assert path == g3.path
        assert name == g3.name

        # create level 3 child group
        g4 = g1.create_group("foo/bar/baz")
        path, name = self._subgroup_path(g1, "foo/bar/baz")
        assert isinstance(g4, Group)
        assert path == g4.path
        assert name == g4.name

        # create level 3 group via root
        g5 = g4.create_group("/a/b/c/")
        assert isinstance(g5, Group)
        assert "a/b/c" == g5.path
        assert "/a/b/c" == g5.name

        # test non-str keys
        class Foo:
            def __init__(self, s):
                self.s = s

            def __str__(self):
                return self.s

        o = Foo("test/object")
        go = g1.create_group(o)
        path, name = self._subgroup_path(g1, str(o))
        assert isinstance(go, Group)
        assert path == go.path
        go = g1.create_group(b"test/bytes")
        path, name = self._subgroup_path(g1, "test/bytes")
        assert isinstance(go, Group)
        assert path == go.path

        # test bad keys
        with pytest.raises(ValueError):
            g1.create_group("foo")  # already exists
        if g1._version == 2:
            with pytest.raises(ValueError):
                g1.create_group("a/b/c")  # already exists
        elif g1._version == 3:
            # for v3 'group/a/b/c' does not already exist
            g1.create_group("a/b/c")
        with pytest.raises(ValueError):
            g4.create_group("/a/b/c")  # already exists
        with pytest.raises(ValueError):
            g1.create_group("")

        # multi
        g6, g7 = g1.create_groups("y", "z")
        assert isinstance(g6, Group)
        assert g6.path == self._subgroup_path(g1, "y")[0]
        assert isinstance(g7, Group)
        assert g7.path == self._subgroup_path(g1, "z")[0]

        g1.store.close()

    def test_require_group(self):
        g1 = self.create_group()

        # test creation
        g2 = g1.require_group("foo")
        path, name = self._subgroup_path(g1, "foo")
        assert isinstance(g2, Group)
        assert path == g2.path
        g3 = g2.require_group("bar")
        path, name = self._subgroup_path(g2, "bar")
        assert isinstance(g3, Group)
        assert path == g3.path
        g4 = g1.require_group("foo/bar/baz")
        path, name = self._subgroup_path(g1, "foo/bar/baz")
        assert isinstance(g4, Group)
        assert path == g4.path
        g5 = g4.require_group("/a/b/c/")
        assert isinstance(g5, Group)
        assert "a/b/c" == g5.path

        # test when already created
        g2a = g1.require_group("foo")
        assert g2 == g2a
        assert g2.store is g2a.store
        g3a = g2a.require_group("bar")
        assert g3 == g3a
        assert g3.store is g3a.store
        g4a = g1.require_group("foo/bar/baz")
        assert g4 == g4a
        assert g4.store is g4a.store
        g5a = g4a.require_group("/a/b/c/")
        assert g5 == g5a
        assert g5.store is g5a.store

        # test path normalization
        if g1._version == 2:
            assert g1.require_group("quux") == g1.require_group("/quux/")
        elif g1._version:
            # These are not equal in v3!
            # 'quux' will be within the group:
            #      meta/root/group/quux.group.json
            # '/quux/' will be outside of the group at:
            #      meta/root/quux.group.json
            assert g1.require_group("quux") != g1.require_group("/quux/")

        # multi
        g6, g7 = g1.require_groups("y", "z")
        assert isinstance(g6, Group)
        assert g6.path == self._subgroup_path(g1, "y")[0]
        assert isinstance(g7, Group)
        assert g7.path == self._subgroup_path(g1, "z")[0]

        g1.store.close()

    def test_rmdir_group_and_array_metadata_files(self):
        """Test group.store's rmdir method.

        This test case was added to complete test coverage of `ABSStore.rmdir`.
        """
        g1 = self.create_group()
        # create a dataset
        g1.create_dataset("arr1", shape=(100,), chunks=(10,), dtype=np.uint8)

        # create level 1 child group
        g2 = g1.create_group("foo")
        g1.create_dataset("arr2", shape=(100,), chunks=(10,), dtype=np.uint8)

        if g1._version > 2 and g1.store.is_erasable():
            arr_path = g1.path + "/arr1"
            sfx = _get_metadata_suffix(g1.store)
            array_meta_file = meta_root + arr_path + ".array" + sfx
            assert array_meta_file in g1.store
            group_meta_file = meta_root + g2.path + ".group" + sfx
            assert group_meta_file in g1.store

            # rmdir on the array path should also remove the metadata file
            g1.store.rmdir(arr_path)
            assert array_meta_file not in g1.store
            # rmdir on the group path should also remove its metadata file
            g1.store.rmdir(g2.path)
            assert group_meta_file not in g1.store

    def _dataset_path(self, group, path):
        path = path.rstrip("/")
        absolute = path.startswith("/")
        if absolute:
            dataset_path = path
        else:
            dataset_path = "/".join([group.path, path])
        dataset_path = dataset_path.lstrip("/")
        dataset_name = "/" + dataset_path
        return dataset_path, dataset_name

    def test_create_dataset(self):
        g = self.create_group()

        # create as immediate child
        dpath = "foo"
        d1 = g.create_dataset(dpath, shape=1000, chunks=100)
        path, name = self._dataset_path(g, dpath)
        assert isinstance(d1, Array)
        assert (1000,) == d1.shape
        assert (100,) == d1.chunks
        assert path == d1.path
        assert name == d1.name
        assert g.store is d1.store

        # create as descendant
        dpath = "/a/b/c/"
        d2 = g.create_dataset(
            dpath,
            shape=2000,
            chunks=200,
            dtype="i1",
            compression="zlib",
            compression_opts=9,
            fill_value=42,
            order="F",
        )
        path, name = self._dataset_path(g, dpath)
        assert isinstance(d2, Array)
        assert (2000,) == d2.shape
        assert (200,) == d2.chunks
        assert np.dtype("i1") == d2.dtype
        assert "zlib" == d2.compressor.codec_id
        assert 9 == d2.compressor.level
        assert 42 == d2.fill_value
        assert "F" == d2.order
        assert path == d2.path
        assert name == d2.name
        assert g.store is d2.store

        # create with data
        data = np.arange(3000, dtype="u2")
        dpath = "bar"
        d3 = g.create_dataset(dpath, data=data, chunks=300)
        path, name = self._dataset_path(g, dpath)
        assert isinstance(d3, Array)
        assert (3000,) == d3.shape
        assert (300,) == d3.chunks
        assert np.dtype("u2") == d3.dtype
        assert_array_equal(data, d3[:])
        assert path == d3.path
        assert name == d3.name
        assert g.store is d3.store

        # compression arguments handling follows...

        # compression_opts as dict
        d = g.create_dataset(
            "aaa",
            shape=1000,
            dtype="u1",
            compression="blosc",
            compression_opts=dict(cname="zstd", clevel=1, shuffle=2),
        )
        assert d.compressor.codec_id == "blosc"
        assert "zstd" == d.compressor.cname
        assert 1 == d.compressor.clevel
        assert 2 == d.compressor.shuffle

        # compression_opts as sequence
        d = g.create_dataset(
            "bbb", shape=1000, dtype="u1", compression="blosc", compression_opts=("zstd", 1, 2)
        )
        assert d.compressor.codec_id == "blosc"
        assert "zstd" == d.compressor.cname
        assert 1 == d.compressor.clevel
        assert 2 == d.compressor.shuffle

        # None compression_opts
        d = g.create_dataset("ccc", shape=1000, dtype="u1", compression="zlib")
        assert d.compressor.codec_id == "zlib"
        assert 1 == d.compressor.level

        # None compression
        d = g.create_dataset("ddd", shape=1000, dtype="u1", compression=None)
        assert d.compressor is None

        # compressor as compression
        d = g.create_dataset("eee", shape=1000, dtype="u1", compression=Zlib(1))
        assert d.compressor.codec_id == "zlib"
        assert 1 == d.compressor.level

        g.store.close()

    def test_require_dataset(self):
        g = self.create_group()

        # create
        dpath = "foo"
        d1 = g.require_dataset(dpath, shape=1000, chunks=100, dtype="f4")
        d1[:] = np.arange(1000)
        path, name = self._dataset_path(g, dpath)
        assert isinstance(d1, Array)
        assert (1000,) == d1.shape
        assert (100,) == d1.chunks
        assert np.dtype("f4") == d1.dtype
        assert path == d1.path
        assert name == d1.name
        assert g.store is d1.store
        assert_array_equal(np.arange(1000), d1[:])

        # require
        d2 = g.require_dataset(dpath, shape=1000, chunks=100, dtype="f4")
        assert isinstance(d2, Array)
        assert (1000,) == d2.shape
        assert (100,) == d2.chunks
        assert np.dtype("f4") == d2.dtype
        assert path == d2.path
        assert name == d2.name
        assert g.store is d2.store
        assert_array_equal(np.arange(1000), d2[:])
        assert d1 == d2

        # bad shape - use TypeError for h5py compatibility
        with pytest.raises(TypeError):
            g.require_dataset("foo", shape=2000, chunks=100, dtype="f4")

        # dtype matching
        # can cast
        d3 = g.require_dataset("foo", shape=1000, chunks=100, dtype="i2")
        assert np.dtype("f4") == d3.dtype
        assert d1 == d3
        with pytest.raises(TypeError):
            # cannot cast
            g.require_dataset("foo", shape=1000, chunks=100, dtype="i4")
        with pytest.raises(TypeError):
            # can cast but not exact match
            g.require_dataset("foo", shape=1000, chunks=100, dtype="i2", exact=True)

        g.store.close()

    def test_create_errors(self):
        g = self.create_group()

        # array obstructs group, array
        g.create_dataset("foo", shape=100, chunks=10)
        with pytest.raises(ValueError):
            g.create_group("foo/bar")
        with pytest.raises(ValueError):
            g.require_group("foo/bar")
        with pytest.raises(ValueError):
            g.create_dataset("foo/bar", shape=100, chunks=10)
        with pytest.raises(ValueError):
            g.require_dataset("foo/bar", shape=100, chunks=10)

        # array obstructs group, array
        g.create_dataset("a/b", shape=100, chunks=10)
        with pytest.raises(ValueError):
            g.create_group("a/b")
        with pytest.raises(ValueError):
            g.require_group("a/b")
        with pytest.raises(ValueError):
            g.create_dataset("a/b", shape=100, chunks=10)

        # group obstructs array
        g.create_group("c/d")
        with pytest.raises(ValueError):
            g.create_dataset("c", shape=100, chunks=10)
        with pytest.raises(ValueError):
            g.require_dataset("c", shape=100, chunks=10)
        with pytest.raises(ValueError):
            g.create_dataset("c/d", shape=100, chunks=10)
        with pytest.raises(ValueError):
            g.require_dataset("c/d", shape=100, chunks=10)

        # h5py compatibility, accept 'fillvalue'
        d = g.create_dataset("x", shape=100, chunks=10, fillvalue=42)
        assert 42 == d.fill_value

        # h5py compatibility, ignore 'shuffle'
        with pytest.warns(UserWarning, match="ignoring keyword argument 'shuffle'"):
            g.create_dataset("y", shape=100, chunks=10, shuffle=True)

        # read-only
        g = self.create_group(read_only=True)
        with pytest.raises(PermissionError):
            g.create_group("zzz")
        with pytest.raises(PermissionError):
            g.require_group("zzz")
        with pytest.raises(PermissionError):
            g.create_dataset("zzz", shape=100, chunks=10)
        with pytest.raises(PermissionError):
            g.require_dataset("zzz", shape=100, chunks=10)

        g.store.close()

    def test_create_overwrite(self):
        try:
            for method_name in "create_dataset", "create", "empty", "zeros", "ones":
                g = self.create_group()
                getattr(g, method_name)("foo", shape=100, chunks=10)

                # overwrite array with array
                d = getattr(g, method_name)("foo", shape=200, chunks=20, overwrite=True)
                assert (200,) == d.shape
                # overwrite array with group
                g2 = g.create_group("foo", overwrite=True)
                assert 0 == len(g2)
                # overwrite group with array
                d = getattr(g, method_name)("foo", shape=300, chunks=30, overwrite=True)
                assert (300,) == d.shape
                # overwrite array with group
                d = getattr(g, method_name)("foo/bar", shape=400, chunks=40, overwrite=True)
                assert (400,) == d.shape
                assert isinstance(g["foo"], Group)

                g.store.close()
        except NotImplementedError:
            pass

    def test_getitem_contains_iterators(self):
        # setup
        g1 = self.create_group()
        g2 = g1.create_group("foo/bar")
        if g1._version == 2:
            d1 = g2.create_dataset("/a/b/c", shape=1000, chunks=100)
        else:
            # v3: cannot create a dataset at the root by starting with /
            #     instead, need to create the dataset on g1 directly
            d1 = g1.create_dataset("a/b/c", shape=1000, chunks=100)
        d1[:] = np.arange(1000)
        d2 = g1.create_dataset("foo/baz", shape=3000, chunks=300)
        d2[:] = np.arange(3000)

        # test __getitem__
        assert isinstance(g1["foo"], Group)
        assert isinstance(g1["foo"]["bar"], Group)
        assert isinstance(g1["foo/bar"], Group)
        if g1._version == 2:
            assert isinstance(g1["/foo/bar/"], Group)
        else:
            # start or end with / raises KeyError
            # TODO: should we allow stripping of these on v3?
            with pytest.raises(KeyError):
                assert isinstance(g1["/foo/bar/"], Group)
        assert isinstance(g1["foo/baz"], Array)
        assert g2 == g1["foo/bar"]
        assert g1["foo"]["bar"] == g1["foo/bar"]
        assert d2 == g1["foo/baz"]
        assert_array_equal(d2[:], g1["foo/baz"])
        assert isinstance(g1["a"], Group)
        assert isinstance(g1["a"]["b"], Group)
        assert isinstance(g1["a/b"], Group)
        assert isinstance(g1["a"]["b"]["c"], Array)
        assert isinstance(g1["a/b/c"], Array)
        assert d1 == g1["a/b/c"]
        assert g1["a"]["b"]["c"] == g1["a/b/c"]
        assert_array_equal(d1[:], g1["a/b/c"][:])

        # test __contains__
        assert "foo" in g1
        assert "foo/bar" in g1
        assert "foo/baz" in g1
        assert "bar" in g1["foo"]
        assert "a" in g1
        assert "a/b" in g1
        assert "a/b/c" in g1
        assert "baz" not in g1
        assert "a/b/c/d" not in g1
        assert "a/z" not in g1
        assert "quux" not in g1["foo"]

        # test key errors
        with pytest.raises(KeyError):
            g1["baz"]
        with pytest.raises(KeyError):
            g1["x/y/z"]

        # test __len__
        assert 2 == len(g1)
        assert 2 == len(g1["foo"])
        assert 0 == len(g1["foo/bar"])
        assert 1 == len(g1["a"])
        assert 1 == len(g1["a/b"])

        # test __iter__, keys()

        if g1._version == 2:
            # currently assumes sorted by key
            assert ["a", "foo"] == list(g1)
            assert ["a", "foo"] == list(g1.keys())
            assert ["bar", "baz"] == list(g1["foo"])
            assert ["bar", "baz"] == list(g1["foo"].keys())
        else:
            # v3 is not necessarily sorted by key
            assert ["a", "foo"] == sorted(list(g1))
            assert ["a", "foo"] == sorted(list(g1.keys()))
            assert ["bar", "baz"] == sorted(list(g1["foo"]))
            assert ["bar", "baz"] == sorted(list(g1["foo"].keys()))
        assert [] == sorted(g1["foo/bar"])
        assert [] == sorted(g1["foo/bar"].keys())

        # test items(), values()
        # currently assumes sorted by key

        items = list(g1.items())
        values = list(g1.values())
        if g1._version == 3:
            # v3 are not automatically sorted by key
            items, values = zip(*sorted(zip(items, values), key=lambda x: x[0]))
        assert "a" == items[0][0]
        assert g1["a"] == items[0][1]
        assert g1["a"] == values[0]
        assert "foo" == items[1][0]
        assert g1["foo"] == items[1][1]
        assert g1["foo"] == values[1]

        items = list(g1["foo"].items())
        values = list(g1["foo"].values())
        if g1._version == 3:
            # v3 are not automatically sorted by key
            items, values = zip(*sorted(zip(items, values), key=lambda x: x[0]))
        assert "bar" == items[0][0]
        assert g1["foo"]["bar"] == items[0][1]
        assert g1["foo"]["bar"] == values[0]
        assert "baz" == items[1][0]
        assert g1["foo"]["baz"] == items[1][1]
        assert g1["foo"]["baz"] == values[1]

        # test array_keys(), arrays(), group_keys(), groups()

        groups = list(g1.groups())
        arrays = list(g1.arrays())
        if g1._version == 2:
            # currently assumes sorted by key
            assert ["a", "foo"] == list(g1.group_keys())
        else:
            assert ["a", "foo"] == sorted(list(g1.group_keys()))
            groups = sorted(groups)
            arrays = sorted(arrays)
        assert "a" == groups[0][0]
        assert g1["a"] == groups[0][1]
        assert "foo" == groups[1][0]
        assert g1["foo"] == groups[1][1]
        assert [] == list(g1.array_keys())
        assert [] == arrays

        assert ["bar"] == list(g1["foo"].group_keys())
        assert ["baz"] == list(g1["foo"].array_keys())
        groups = list(g1["foo"].groups())
        arrays = list(g1["foo"].arrays())
        if g1._version == 3:
            groups = sorted(groups)
            arrays = sorted(arrays)
        assert "bar" == groups[0][0]
        assert g1["foo"]["bar"] == groups[0][1]
        assert "baz" == arrays[0][0]
        assert g1["foo"]["baz"] == arrays[0][1]

        # visitor collection tests
        items = []

        def visitor2(obj):
            items.append(obj.path)

        # noinspection PyUnusedLocal
        def visitor3(name, obj=None):
            items.append(name)

        def visitor4(name, obj):
            items.append((name, obj))

        del items[:]
        g1.visitvalues(visitor2)
        expected_items = [
            "a",
            "a/b",
            "a/b/c",
            "foo",
            "foo/bar",
            "foo/baz",
        ]
        if g1._version == 3:
            expected_items = [g1.path + "/" + i for i in expected_items]
        assert expected_items == items

        del items[:]
        g1["foo"].visitvalues(visitor2)
        expected_items = [
            "foo/bar",
            "foo/baz",
        ]
        if g1._version == 3:
            expected_items = [g1.path + "/" + i for i in expected_items]
        assert expected_items == items

        del items[:]
        g1.visit(visitor3)
        assert [
            "a",
            "a/b",
            "a/b/c",
            "foo",
            "foo/bar",
            "foo/baz",
        ] == items

        del items[:]
        g1["foo"].visit(visitor3)
        assert [
            "bar",
            "baz",
        ] == items

        del items[:]
        g1.visitkeys(visitor3)
        assert [
            "a",
            "a/b",
            "a/b/c",
            "foo",
            "foo/bar",
            "foo/baz",
        ] == items

        del items[:]
        g1["foo"].visitkeys(visitor3)
        assert [
            "bar",
            "baz",
        ] == items

        del items[:]
        g1.visititems(visitor3)
        assert [
            "a",
            "a/b",
            "a/b/c",
            "foo",
            "foo/bar",
            "foo/baz",
        ] == items

        del items[:]
        g1["foo"].visititems(visitor3)
        assert [
            "bar",
            "baz",
        ] == items

        del items[:]
        g1.visititems(visitor4)
        for n, o in items:
            assert g1[n] == o

        del items[:]
        g1["foo"].visititems(visitor4)
        for n, o in items:
            assert g1["foo"][n] == o

        # visitor filter tests
        # noinspection PyUnusedLocal
        def visitor0(val, *args):
            name = getattr(val, "path", val)
            if name == "a/b/c/d":
                return True  # pragma: no cover

        # noinspection PyUnusedLocal
        def visitor1(val, *args):
            name = getattr(val, "path", val)
            if name.startswith("group/"):
                # strip the group path for v3
                name = name[6:]
            if name == "a/b/c":
                return True

        assert g1.visit(visitor0) is None
        assert g1.visitkeys(visitor0) is None
        assert g1.visitvalues(visitor0) is None
        assert g1.visititems(visitor0) is None
        assert g1.visit(visitor1) is True
        assert g1.visitkeys(visitor1) is True
        assert g1.visitvalues(visitor1) is True
        assert g1.visititems(visitor1) is True

        g1.store.close()

    # regression test for https://github.com/zarr-developers/zarr-python/issues/1228
    def test_double_counting_group_v3(self):
        root_group = self.create_group()
        group_names = ["foo", "foo-", "foo_"]
        for name in group_names:
            sub_group = root_group.create_group(name)
            sub_group.create("bar", shape=10, dtype="i4")
        assert list(root_group.group_keys()) == sorted(group_names)
        assert list(root_group.groups()) == [
            (name, root_group[name]) for name in sorted(group_names)
        ]

    def test_empty_getitem_contains_iterators(self):
        # setup
        g = self.create_group()

        # test
        assert [] == list(g)
        assert [] == list(g.keys())
        assert 0 == len(g)
        assert "foo" not in g

        g.store.close()

    def test_iterators_recurse(self):
        # setup
        g1 = self.create_group()
        g2 = g1.create_group("foo/bar")
        d1 = g2.create_dataset("/a/b/c", shape=1000, chunks=100)
        d1[:] = np.arange(1000)
        d2 = g1.create_dataset("foo/baz", shape=3000, chunks=300)
        d2[:] = np.arange(3000)
        d3 = g2.create_dataset("zab", shape=2000, chunks=200)
        d3[:] = np.arange(2000)

        # test recursive array_keys
        array_keys = list(g1["foo"].array_keys(recurse=False))
        array_keys_recurse = list(g1["foo"].array_keys(recurse=True))
        assert len(array_keys_recurse) > len(array_keys)
        assert sorted(array_keys_recurse) == ["baz", "zab"]

        # test recursive arrays
        arrays = list(g1["foo"].arrays(recurse=False))
        arrays_recurse = list(g1["foo"].arrays(recurse=True))
        assert len(arrays_recurse) > len(arrays)
        assert "zab" == arrays_recurse[0][0]
        assert g1["foo"]["bar"]["zab"] == arrays_recurse[0][1]

        g1.store.close()

    def test_getattr(self):
        # setup
        g1 = self.create_group()
        g2 = g1.create_group("foo")
        g2.create_dataset("bar", shape=100)

        # test
        assert g1["foo"] == g1.foo
        assert g2["bar"] == g2.bar
        # test that hasattr returns False instead of an exception (issue #88)
        assert not hasattr(g1, "unexistingattribute")

        g1.store.close()

    def test_setitem(self):
        g = self.create_group()
        try:
            data = np.arange(100)
            g["foo"] = data
            assert_array_equal(data, g["foo"])
            data = np.arange(200)
            g["foo"] = data
            assert_array_equal(data, g["foo"])
            # 0d array
            g["foo"] = 42
            assert () == g["foo"].shape
            assert 42 == g["foo"][()]
        except NotImplementedError:
            pass
        g.store.close()

    def test_delitem(self):
        g = self.create_group()
        g.create_group("foo")
        g.create_dataset("bar/baz", shape=100, chunks=10)
        assert "foo" in g
        assert "bar" in g
        assert "bar/baz" in g
        try:
            del g["bar"]
            with pytest.raises(KeyError):
                del g["xxx"]
        except NotImplementedError:
            pass
        else:
            assert "foo" in g
            assert "bar" not in g
            assert "bar/baz" not in g
        g.store.close()

    def test_move(self):
        g = self.create_group()

        data = np.arange(100)
        g["boo"] = data

        data = np.arange(100)
        g["foo"] = data

        g.move("foo", "bar")
        assert "foo" not in g
        assert "bar" in g
        assert_array_equal(data, g["bar"])

        g.move("bar", "foo/bar")
        assert "bar" not in g
        assert "foo" in g
        assert "foo/bar" in g
        assert isinstance(g["foo"], Group)
        assert_array_equal(data, g["foo/bar"])

        g.move("foo", "foo2")
        assert "foo" not in g
        assert "foo/bar" not in g
        assert "foo2" in g
        assert "foo2/bar" in g
        assert isinstance(g["foo2"], Group)
        assert_array_equal(data, g["foo2/bar"])

        g2 = g["foo2"]
        g2.move("bar", "/bar")
        assert "foo2" in g
        assert "foo2/bar" not in g
        if g2._version == 2:
            assert "bar" in g
        else:
            # The `g2.move` call above moved bar to meta/root/bar and
            # meta/data/bar. This is outside the `g` group located at
            # /meta/root/group, so bar is no longer within `g`.
            assert "bar" not in g
            assert "meta/root/bar.array.json" in g._store
            if g._chunk_store:
                assert "data/root/bar/c0" in g._chunk_store
            else:
                assert "data/root/bar/c0" in g._store
        assert isinstance(g["foo2"], Group)
        if g2._version == 2:
            assert_array_equal(data, g["bar"])
        else:
            # TODO: How to access element created outside of group.path in v3?
            #       One option is to make a Hierarchy class representing the
            #       root. Currently Group requires specification of `path`,
            #       but the path of the root would be just '' which is not
            #       currently allowed.
            pass

        with pytest.raises(ValueError):
            g2.move("bar", "bar2")

        with pytest.raises(ValueError):
            g.move("bar", "boo")

        g.store.close()

    def test_array_creation(self):
        grp = self.create_group()

        a = grp.create("a", shape=100, chunks=10)
        assert isinstance(a, Array)
        b = grp.empty("b", shape=100, chunks=10)
        assert isinstance(b, Array)
        assert b.fill_value is None
        c = grp.zeros("c", shape=100, chunks=10)
        assert isinstance(c, Array)
        assert 0 == c.fill_value
        d = grp.ones("d", shape=100, chunks=10)
        assert isinstance(d, Array)
        assert 1 == d.fill_value
        e = grp.full("e", shape=100, chunks=10, fill_value=42)
        assert isinstance(e, Array)
        assert 42 == e.fill_value

        f = grp.empty_like("f", a)
        assert isinstance(f, Array)
        assert f.fill_value is None
        g = grp.zeros_like("g", a)
        assert isinstance(g, Array)
        assert 0 == g.fill_value
        h = grp.ones_like("h", a)
        assert isinstance(h, Array)
        assert 1 == h.fill_value
        i = grp.full_like("i", e)
        assert isinstance(i, Array)
        assert 42 == i.fill_value

        j = grp.array("j", data=np.arange(100), chunks=10)
        assert isinstance(j, Array)
        assert_array_equal(np.arange(100), j[:])

        grp.store.close()

        grp = self.create_group(read_only=True)
        with pytest.raises(PermissionError):
            grp.create("aa", shape=100, chunks=10)
        with pytest.raises(PermissionError):
            grp.empty("aa", shape=100, chunks=10)
        with pytest.raises(PermissionError):
            grp.zeros("aa", shape=100, chunks=10)
        with pytest.raises(PermissionError):
            grp.ones("aa", shape=100, chunks=10)
        with pytest.raises(PermissionError):
            grp.full("aa", shape=100, chunks=10, fill_value=42)
        with pytest.raises(PermissionError):
            grp.array("aa", data=np.arange(100), chunks=10)
        with pytest.raises(PermissionError):
            grp.create("aa", shape=100, chunks=10)
        with pytest.raises(PermissionError):
            grp.empty_like("aa", a)
        with pytest.raises(PermissionError):
            grp.zeros_like("aa", a)
        with pytest.raises(PermissionError):
            grp.ones_like("aa", a)
        with pytest.raises(PermissionError):
            grp.full_like("aa", a)

        grp.store.close()

    def test_paths(self):
        g1 = self.create_group()
        g2 = g1.create_group("foo/bar")

        if g1._version == 2:
            assert g1 == g1["/"]
            assert g1 == g1["//"]
            assert g1 == g1["///"]
            assert g1 == g2["/"]
            assert g1 == g2["//"]
            assert g1 == g2["///"]
            assert g2 == g1["foo/bar"]
            assert g2 == g1["/foo/bar"]
            assert g2 == g1["foo/bar/"]
            assert g2 == g1["//foo/bar"]
            assert g2 == g1["//foo//bar//"]
            assert g2 == g1["///foo///bar///"]
            assert g2 == g2["/foo/bar"]
        else:
            # the expected key format gives a match
            assert g2 == g1["foo/bar"]

            # TODO: Should presence of a trailing slash raise KeyError?
            # The spec says "the final character is not a / character"
            # but we currently strip trailing '/' as done for v2.
            assert g2 == g1["foo/bar/"]

            # double slash also currently works (spec doesn't mention this
            # case, but have kept it for v2 behavior compatibility)
            assert g2 == g1["foo//bar"]

            # TODO, root: fix these cases
            # v3: leading / implies we are at the root, not within a group,
            # so these all raise KeyError
            for path in ["/foo/bar", "//foo/bar", "//foo//bar//", "///fooo///bar///"]:
                with pytest.raises(KeyError):
                    g1[path]

        with pytest.raises(ValueError):
            g1["."]
        with pytest.raises(ValueError):
            g1[".."]
        with pytest.raises(ValueError):
            g1["foo/."]
        with pytest.raises(ValueError):
            g1["foo/.."]
        with pytest.raises(ValueError):
            g1["foo/./bar"]
        with pytest.raises(ValueError):
            g1["foo/../bar"]

        g1.store.close()

    def test_pickle(self):
        # setup group
        g = self.create_group()
        d = g.create_dataset("foo/bar", shape=100, chunks=10)
        d[:] = np.arange(100)
        path = g.path
        name = g.name
        n = len(g)
        keys = list(g)

        # round-trip through pickle
        dump = pickle.dumps(g)
        # some stores cannot be opened twice at the same time, need to close
        # store before can round-trip through pickle
        g.store.close()
        g2 = pickle.loads(dump)

        # verify
        assert path == g2.path
        assert name == g2.name
        assert n == len(g2)
        assert keys == list(g2)
        assert isinstance(g2["foo"], Group)
        assert isinstance(g2["foo/bar"], Array)

        g2.store.close()

    def test_context_manager(self):
        with self.create_group() as g:
            d = g.create_dataset("foo/bar", shape=100, chunks=10)
            d[:] = np.arange(100)


@pytest.mark.parametrize("chunk_dict", [False, True])
def test_group_init_from_dict(chunk_dict):
    if chunk_dict:
        store, chunk_store = dict(), dict()
    else:
        store, chunk_store = dict(), None
    init_group(store, path=None, chunk_store=chunk_store)
    g = Group(store, path=None, read_only=False, chunk_store=chunk_store)
    assert store is not g.store
    assert isinstance(g.store, KVStore)
    if chunk_store is None:
        assert g.store is g.chunk_store
    else:
        assert chunk_store is not g.chunk_store


# noinspection PyStatementEffect
@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
class TestGroupV3(TestGroup, unittest.TestCase):
    @staticmethod
    def create_store():
        # can be overridden in sub-classes
        return KVStoreV3(dict()), None

    def create_group(
        self, store=None, path="group", read_only=False, chunk_store=None, synchronizer=None
    ):
        # can be overridden in sub-classes
        if store is None:
            store, chunk_store = self.create_store()
        init_group(store, path=path, chunk_store=chunk_store)
        g = Group(
            store,
            path=path,
            read_only=read_only,
            chunk_store=chunk_store,
            synchronizer=synchronizer,
        )
        return g

    def test_group_init_1(self):
        store, chunk_store = self.create_store()
        g = self.create_group(store, chunk_store=chunk_store)
        assert store is g.store
        if chunk_store is None:
            assert store is g.chunk_store
        else:
            assert chunk_store is g.chunk_store
        assert not g.read_only
        # different path/name in v3 case
        assert "group" == g.path
        assert "/group" == g.name
        assert "group" == g.basename

        assert isinstance(g.attrs, Attributes)
        g.attrs["foo"] = "bar"
        assert g.attrs["foo"] == "bar"

        assert isinstance(g.info, InfoReporter)
        assert isinstance(repr(g.info), str)
        assert isinstance(g.info._repr_html_(), str)
        store.close()

    def test_group_init_errors_2(self):
        store, chunk_store = self.create_store()
        path = "tmp"
        init_array(store, path=path, shape=1000, chunks=100, chunk_store=chunk_store)
        # array blocks group
        with pytest.raises(ValueError):
            Group(store, path=path, chunk_store=chunk_store)
        store.close()


class TestGroupWithMemoryStore(TestGroup):
    @staticmethod
    def create_store():
        return MemoryStore(), None


# noinspection PyStatementEffect
@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
class TestGroupV3WithMemoryStore(TestGroupWithMemoryStore, TestGroupV3):
    @staticmethod
    def create_store():
        return MemoryStoreV3(), None


class TestGroupWithDirectoryStore(TestGroup):
    @staticmethod
    def create_store():
        path = tempfile.mkdtemp()
        atexit.register(atexit_rmtree, path)
        store = DirectoryStore(path)
        return store, None


@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
class TestGroupV3WithDirectoryStore(TestGroupWithDirectoryStore, TestGroupV3):
    @staticmethod
    def create_store():
        path = tempfile.mkdtemp()
        atexit.register(atexit_rmtree, path)
        store = DirectoryStoreV3(path)
        return store, None


@skip_test_env_var("ZARR_TEST_ABS")
class TestGroupWithABSStore(TestGroup):
    @staticmethod
    def create_store():
        container_client = abs_container()
        store = ABSStore(client=container_client)
        store.rmdir()
        return store, None

    @pytest.mark.skipif(sys.version_info < (3, 7), reason="attr not serializable in py36")
    def test_pickle(self):
        # internal attribute on ContainerClient isn't serializable for py36 and earlier
        super().test_pickle()


@skip_test_env_var("ZARR_TEST_ABS")
@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
class TestGroupV3WithABSStore(TestGroupV3):
    @staticmethod
    def create_store():
        container_client = abs_container()
        store = ABSStoreV3(client=container_client)
        store.rmdir()
        return store, None

    @pytest.mark.skipif(sys.version_info < (3, 7), reason="attr not serializable in py36")
    def test_pickle(self):
        # internal attribute on ContainerClient isn't serializable for py36 and earlier
        super().test_pickle()


class TestGroupWithNestedDirectoryStore(TestGroup):
    @staticmethod
    def create_store():
        path = tempfile.mkdtemp()
        atexit.register(atexit_rmtree, path)
        store = NestedDirectoryStore(path)
        return store, None


@pytest.mark.skipif(have_fsspec is False, reason="needs fsspec")
class TestGroupWithFSStore(TestGroup):
    @staticmethod
    def create_store():
        path = tempfile.mkdtemp()
        atexit.register(atexit_rmtree, path)
        store = FSStore(path)
        return store, None

    def test_round_trip_nd(self):
        data = np.arange(1000).reshape(10, 10, 10)
        name = "raw"

        store, _ = self.create_store()
        f = open_group(store, mode="w")
        f.create_dataset(name, data=data, chunks=(5, 5, 5), compressor=None)
        assert name in f
        h = open_group(store, mode="r")
        np.testing.assert_array_equal(h[name][:], data)


@pytest.mark.skipif(have_fsspec is False, reason="needs fsspec")
@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
class TestGroupV3WithFSStore(TestGroupWithFSStore, TestGroupV3):
    @staticmethod
    def create_store():
        path = tempfile.mkdtemp()
        atexit.register(atexit_rmtree, path)
        store = FSStoreV3(path)
        return store, None

    def test_round_trip_nd(self):
        data = np.arange(1000).reshape(10, 10, 10)
        name = "raw"

        store, _ = self.create_store()
        f = open_group(store, path="group", mode="w")
        f.create_dataset(name, data=data, chunks=(5, 5, 5), compressor=None)
        h = open_group(store, path="group", mode="r")
        np.testing.assert_array_equal(h[name][:], data)

        f = open_group(store, path="group2", mode="w")

        data_size = data.nbytes
        group_meta_size = buffer_size(store[meta_root + "group.group.json"])
        group2_meta_size = buffer_size(store[meta_root + "group2.group.json"])
        array_meta_size = buffer_size(store[meta_root + "group/raw.array.json"])
        assert store.getsize() == data_size + group_meta_size + group2_meta_size + array_meta_size
        # added case with path to complete coverage
        assert store.getsize("group") == data_size + group_meta_size + array_meta_size
        assert store.getsize("group2") == group2_meta_size
        assert store.getsize("group/raw") == data_size + array_meta_size


@pytest.mark.skipif(have_fsspec is False, reason="needs fsspec")
class TestGroupWithNestedFSStore(TestGroupWithFSStore):
    @staticmethod
    def create_store():
        path = tempfile.mkdtemp()
        atexit.register(atexit_rmtree, path)
        store = FSStore(path, key_separator="/", auto_mkdir=True)
        return store, None

    def test_inconsistent_dimension_separator(self):
        data = np.arange(1000).reshape(10, 10, 10)
        name = "raw"

        store, _ = self.create_store()
        f = open_group(store, mode="w")

        # cannot specify dimension_separator that conflicts with the store
        with pytest.raises(ValueError):
            f.create_dataset(
                name, data=data, chunks=(5, 5, 5), compressor=None, dimension_separator="."
            )


@pytest.mark.skipif(have_fsspec is False, reason="needs fsspec")
@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
class TestGroupV3WithNestedFSStore(TestGroupV3WithFSStore):
    @staticmethod
    def create_store():
        path = tempfile.mkdtemp()
        atexit.register(atexit_rmtree, path)
        store = FSStoreV3(path, key_separator="/", auto_mkdir=True)
        return store, None

    def test_inconsistent_dimension_separator(self):
        data = np.arange(1000).reshape(10, 10, 10)
        name = "raw"

        store, _ = self.create_store()
        f = open_group(store, path="group", mode="w")

        # cannot specify dimension_separator that conflicts with the store
        with pytest.raises(ValueError):
            f.create_dataset(
                name, data=data, chunks=(5, 5, 5), compressor=None, dimension_separator="."
            )


class TestGroupWithZipStore(TestGroup):
    @staticmethod
    def create_store():
        path = mktemp(suffix=".zip")
        atexit.register(os.remove, path)
        store = ZipStore(path)
        return store, None

    def test_context_manager(self):
        with self.create_group() as g:
            store = g.store
            d = g.create_dataset("foo/bar", shape=100, chunks=10)
            d[:] = np.arange(100)

        # Check that exiting the context manager closes the store,
        # and therefore the underlying ZipFile.
        with pytest.raises(ValueError):
            store.zf.extractall()

    def test_move(self):
        # zip store is not erasable (can so far only append to a zip
        # so we can't test for move.
        pass


@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
class TestGroupV3WithZipStore(TestGroupWithZipStore, TestGroupV3):
    @staticmethod
    def create_store():
        path = mktemp(suffix=".zip")
        atexit.register(os.remove, path)
        store = ZipStoreV3(path)
        return store, None


class TestGroupWithDBMStore(TestGroup):
    @staticmethod
    def create_store():
        path = mktemp(suffix=".anydbm")
        atexit.register(atexit_rmglob, path + "*")
        store = DBMStore(path, flag="n")
        return store, None


@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
class TestGroupV3WithDBMStore(TestGroupWithDBMStore, TestGroupV3):
    @staticmethod
    def create_store():
        path = mktemp(suffix=".anydbm")
        atexit.register(atexit_rmglob, path + "*")
        store = DBMStoreV3(path, flag="n")
        return store, None


class TestGroupWithLMDBStore(TestGroup):
    @staticmethod
    def create_store():
        pytest.importorskip("lmdb")
        path = mktemp(suffix=".lmdb")
        atexit.register(atexit_rmtree, path)
        store = LMDBStore(path)
        return store, None


@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
class TestGroupV3WithLMDBStore(TestGroupWithLMDBStore, TestGroupV3):
    @staticmethod
    def create_store():
        pytest.importorskip("lmdb")
        path = mktemp(suffix=".lmdb")
        atexit.register(atexit_rmtree, path)
        store = LMDBStoreV3(path)
        return store, None


class TestGroupWithSQLiteStore(TestGroup):
    def create_store(self):
        pytest.importorskip("sqlite3")
        path = mktemp(suffix=".db")
        atexit.register(atexit_rmtree, path)
        store = SQLiteStore(path)
        return store, None


@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
class TestGroupV3WithSQLiteStore(TestGroupWithSQLiteStore, TestGroupV3):
    def create_store(self):
        pytest.importorskip("sqlite3")
        path = mktemp(suffix=".db")
        atexit.register(atexit_rmtree, path)
        store = SQLiteStoreV3(path)
        return store, None


class TestGroupWithChunkStore(TestGroup):
    @staticmethod
    def create_store():
        return KVStore(dict()), KVStore(dict())

    def test_chunk_store(self):
        # setup
        store, chunk_store = self.create_store()
        g = self.create_group(store, chunk_store=chunk_store)

        # check attributes
        assert store is g.store
        assert chunk_store is g.chunk_store

        # create array
        a = g.zeros("foo", shape=100, chunks=10)
        assert store is a.store
        assert chunk_store is a.chunk_store
        a[:] = np.arange(100)
        assert_array_equal(np.arange(100), a[:])

        # check store keys
        expect = sorted([group_meta_key, "foo/" + array_meta_key])
        actual = sorted(store.keys())
        assert expect == actual
        expect = ["foo/" + str(i) for i in range(10)]
        actual = sorted(chunk_store.keys())
        assert expect == actual


@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
class TestGroupV3WithChunkStore(TestGroupWithChunkStore, TestGroupV3):
    @staticmethod
    def create_store():
        return KVStoreV3(dict()), KVStoreV3(dict())

    def test_chunk_store(self):
        # setup
        store, chunk_store = self.create_store()
        path = "group1"
        g = self.create_group(store, path=path, chunk_store=chunk_store)

        # check attributes
        assert store is g.store
        assert chunk_store is g.chunk_store

        # create array
        a = g.zeros("foo", shape=100, chunks=10)
        assert store is a.store
        assert chunk_store is a.chunk_store
        a[:] = np.arange(100)
        assert_array_equal(np.arange(100), a[:])

        # check store keys
        group_key = meta_root + path + ".group.json"
        array_key = meta_root + path + "/foo" + ".array.json"
        expect = sorted([group_key, array_key, "zarr.json"])
        actual = sorted(store.keys())
        assert expect == actual
        expect = [data_root + path + "/foo/c" + str(i) for i in range(10)]
        expect += ["zarr.json"]
        actual = sorted(chunk_store.keys())
        assert expect == actual


class TestGroupWithStoreCache(TestGroup):
    @staticmethod
    def create_store():
        store = LRUStoreCache(dict(), max_size=None)
        return store, None


@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
class TestGroupV3WithStoreCache(TestGroupWithStoreCache, TestGroupV3):
    @staticmethod
    def create_store():
        store = LRUStoreCacheV3(dict(), max_size=None)
        return store, None


@pytest.mark.parametrize("zarr_version", _VERSIONS)
def test_group(zarr_version):
    # test the group() convenience function

    # basic usage
    if zarr_version == 2:
        g = group()
        assert "" == g.path
        assert "/" == g.name
    else:
        g = group(path="group1", zarr_version=zarr_version)
        assert "group1" == g.path
        assert "/group1" == g.name
    assert isinstance(g, Group)

    # usage with custom store
    if zarr_version == 2:
        store = KVStore(dict())
        path = None
    else:
        store = KVStoreV3(dict())
        path = "foo"
    g = group(store=store, path=path)
    assert isinstance(g, Group)
    assert store is g.store

    # overwrite behaviour
    if zarr_version == 2:
        store = KVStore(dict())
        path = None
    else:
        store = KVStoreV3(dict())
        path = "foo"
    init_array(store, path=path, shape=100, chunks=10)
    with pytest.raises(ValueError):
        group(store, path=path)
    g = group(store, path=path, overwrite=True)
    assert isinstance(g, Group)
    assert store is g.store


@pytest.mark.skipif(have_fsspec is False, reason="needs fsspec")
@pytest.mark.parametrize("zarr_version", _VERSIONS)
def test_group_writeable_mode(zarr_version, tmp_path):
    # Regression test for https://github.com/zarr-developers/zarr-python/issues/1353
    import fsspec

    store = fsspec.get_mapper(str(tmp_path))
    zg = group(store=store)
    assert zg.store.map == store


@pytest.mark.parametrize("zarr_version", _VERSIONS)
def test_open_group(zarr_version):
    # test the open_group() convenience function

    store = "data/group.zarr"

    expected_store_type = DirectoryStore if zarr_version == 2 else DirectoryStoreV3

    # mode == 'w'
    path = None if zarr_version == 2 else "group1"
    g = open_group(store, path=path, mode="w", zarr_version=zarr_version)
    assert isinstance(g, Group)
    assert isinstance(g.store, expected_store_type)
    assert 0 == len(g)
    g.create_groups("foo", "bar")
    assert 2 == len(g)

    # mode in 'r', 'r+'
    open_array("data/array.zarr", shape=100, chunks=10, mode="w")
    for mode in "r", "r+":
        with pytest.raises(ValueError):
            open_group("doesnotexist", mode=mode)
        with pytest.raises(ValueError):
            open_group("data/array.zarr", mode=mode)
    g = open_group(store, mode="r")
    assert isinstance(g, Group)
    assert 2 == len(g)
    with pytest.raises(PermissionError):
        g.create_group("baz")
    g = open_group(store, mode="r+")
    assert isinstance(g, Group)
    assert 2 == len(g)
    g.create_groups("baz", "quux")
    assert 4 == len(g)

    # mode == 'a'
    shutil.rmtree(store)
    g = open_group(store, path=path, mode="a", zarr_version=zarr_version)
    assert isinstance(g, Group)
    assert isinstance(g.store, expected_store_type)
    assert 0 == len(g)
    g.create_groups("foo", "bar")
    assert 2 == len(g)
    if zarr_version == 2:
        with pytest.raises(ValueError):
            open_group("data/array.zarr", mode="a", zarr_version=zarr_version)
    else:
        # TODO, root: should this raise an error?
        open_group("data/array.zarr", mode="a", zarr_version=zarr_version)

    # mode in 'w-', 'x'
    for mode in "w-", "x":
        shutil.rmtree(store)
        g = open_group(store, path=path, mode=mode, zarr_version=zarr_version)
        assert isinstance(g, Group)
        assert isinstance(g.store, expected_store_type)
        assert 0 == len(g)
        g.create_groups("foo", "bar")
        assert 2 == len(g)
        with pytest.raises(ValueError):
            open_group(store, path=path, mode=mode, zarr_version=zarr_version)
        if zarr_version == 2:
            with pytest.raises(ValueError):
                open_group("data/array.zarr", mode=mode)

    # open with path
    g = open_group(store, path="foo/bar", zarr_version=zarr_version)
    assert isinstance(g, Group)
    assert "foo/bar" == g.path


@pytest.mark.parametrize("zarr_version", _VERSIONS)
def test_group_completions(zarr_version):
    path = None if zarr_version == 2 else "group1"
    g = group(path=path, zarr_version=zarr_version)
    d = dir(g)
    assert "foo" not in d
    assert "bar" not in d
    assert "baz" not in d
    assert "qux" not in d
    assert "xxx" not in d
    assert "yyy" not in d
    assert "zzz" not in d
    assert "123" not in d
    assert "456" not in d
    g.create_groups("foo", "bar", "baz/qux", "123")
    g.zeros("xxx", shape=100)
    g.zeros("yyy", shape=100)
    g.zeros("zzz", shape=100)
    g.zeros("456", shape=100)
    d = dir(g)
    assert "foo" in d
    assert "bar" in d
    assert "baz" in d
    assert "qux" not in d
    assert "xxx" in d
    assert "yyy" in d
    assert "zzz" in d
    assert "123" not in d  # not valid identifier
    assert "456" not in d  # not valid identifier


@pytest.mark.parametrize("zarr_version", _VERSIONS)
def test_group_key_completions(zarr_version):
    path = None if zarr_version == 2 else "group1"
    g = group(path=path, zarr_version=zarr_version)
    d = dir(g)
    # noinspection PyProtectedMember
    k = g._ipython_key_completions_()

    # none of these names should be an attribute
    assert "foo" not in d
    assert "bar" not in d
    assert "baz" not in d
    assert "qux" not in d
    assert "xxx" not in d
    assert "yyy" not in d
    assert "zzz" not in d
    assert "123" not in d
    assert "456" not in d
    assert "asdf;" not in d

    # none of these names should be an item
    assert "foo" not in k
    assert "bar" not in k
    assert "baz" not in k
    assert "qux" not in k
    assert "xxx" not in k
    assert "yyy" not in k
    assert "zzz" not in k
    assert "123" not in k
    assert "456" not in k
    assert "asdf;" not in k

    g.create_groups("foo", "bar", "baz/qux", "123")
    g.zeros("xxx", shape=100)
    g.zeros("yyy", shape=100)
    g.zeros("zzz", shape=100)
    g.zeros("456", shape=100)
    if zarr_version == 2:
        g.zeros("asdf;", shape=100)
    else:
        # cannot have ; in key name for v3
        with pytest.raises(ValueError):
            g.zeros("asdf;", shape=100)

    d = dir(g)
    # noinspection PyProtectedMember
    k = g._ipython_key_completions_()

    assert "foo" in d
    assert "bar" in d
    assert "baz" in d
    assert "qux" not in d
    assert "xxx" in d
    assert "yyy" in d
    assert "zzz" in d
    assert "123" not in d  # not valid identifier
    assert "456" not in d  # not valid identifier
    if zarr_version == 2:
        assert "asdf;" not in d  # not valid identifier

    assert "foo" in k
    assert "bar" in k
    assert "baz" in k
    assert "qux" not in k
    assert "xxx" in k
    assert "yyy" in k
    assert "zzz" in k
    assert "123" in k
    assert "456" in k
    if zarr_version == 2:
        assert "asdf;" in k


def _check_tree(g, expect_bytes, expect_text):
    assert expect_bytes == bytes(g.tree())
    assert expect_text == str(g.tree())
    expect_repr = expect_text
    assert expect_repr == repr(g.tree())
    if ipytree:
        # noinspection PyProtectedMember
        widget = g.tree()._repr_mimebundle_()
        isinstance(widget, ipytree.Tree)


@pytest.mark.parametrize("zarr_version", _VERSIONS)
@pytest.mark.parametrize("at_root", [False, True])
def test_tree(zarr_version, at_root):
    # setup
    path = None if at_root else "group1"
    g1 = group(path=path, zarr_version=zarr_version)
    g2 = g1.create_group("foo")
    g3 = g1.create_group("bar")
    g3.create_group("baz")
    g5 = g3.create_group("quux")
    g5.create_dataset("baz", shape=100, chunks=10)

    tree_path = "/" if at_root else path
    # test root group
    if zarr_version == 2:
        expect_bytes = textwrap.dedent(
            f"""\
        {tree_path}
         +-- bar
         |   +-- baz
         |   +-- quux
         |       +-- baz (100,) float64
         +-- foo"""
        ).encode()
        expect_text = textwrap.dedent(
            f"""\
        {tree_path}
         ├── bar
         │   ├── baz
         │   └── quux
         │       └── baz (100,) float64
         └── foo"""
        )
    else:
        # Almost the same as for v2, but has a path name and the
        # subgroups are not necessarily sorted alphabetically.
        expect_bytes = textwrap.dedent(
            f"""\
        {tree_path}
         +-- foo
         +-- bar
             +-- baz
             +-- quux
                 +-- baz (100,) float64"""
        ).encode()
        expect_text = textwrap.dedent(
            f"""\
        {tree_path}
         ├── foo
         └── bar
             ├── baz
             └── quux
                 └── baz (100,) float64"""
        )
    _check_tree(g1, expect_bytes, expect_text)

    # test different group
    expect_bytes = textwrap.dedent(
        """\
    foo"""
    ).encode()
    expect_text = textwrap.dedent(
        """\
    foo"""
    )
    _check_tree(g2, expect_bytes, expect_text)

    # test different group
    expect_bytes = textwrap.dedent(
        """\
    bar
     +-- baz
     +-- quux
         +-- baz (100,) float64"""
    ).encode()
    expect_text = textwrap.dedent(
        """\
    bar
     ├── baz
     └── quux
         └── baz (100,) float64"""
    )
    _check_tree(g3, expect_bytes, expect_text)


@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
def test_group_mismatched_store_versions():
    store_v3 = KVStoreV3(dict())
    store_v2 = KVStore(dict())

    # separate chunk store
    chunk_store_v2 = KVStore(dict())
    chunk_store_v3 = KVStoreV3(dict())

    init_group(store_v2, path="group1", chunk_store=chunk_store_v2)
    init_group(store_v3, path="group1", chunk_store=chunk_store_v3)

    g1_v3 = Group(store_v3, path="group1", read_only=True, chunk_store=chunk_store_v3)
    assert isinstance(g1_v3._store, KVStoreV3)
    g1_v2 = Group(store_v2, path="group1", read_only=True, chunk_store=chunk_store_v2)
    assert isinstance(g1_v2._store, KVStore)

    # store and chunk_store must have the same zarr protocol version
    with pytest.raises(ValueError):
        Group(store_v3, path="group1", read_only=False, chunk_store=chunk_store_v2)
    with pytest.raises(ValueError):
        Group(store_v2, path="group1", read_only=False, chunk_store=chunk_store_v3)
    with pytest.raises(ValueError):
        open_group(store_v2, path="group1", chunk_store=chunk_store_v3)
    with pytest.raises(ValueError):
        open_group(store_v3, path="group1", chunk_store=chunk_store_v2)

    # raises Value if read_only and path is not a pre-existing group
    with pytest.raises(ValueError):
        Group(store_v3, path="group2", read_only=True, chunk_store=chunk_store_v3)
    with pytest.raises(ValueError):
        Group(store_v3, path="group2", read_only=True, chunk_store=chunk_store_v3)


@pytest.mark.parametrize("zarr_version", _VERSIONS)
def test_open_group_from_paths(zarr_version):
    """Verify zarr_version is applied to both the store and chunk_store."""
    store = tempfile.mkdtemp()
    chunk_store = tempfile.mkdtemp()
    atexit.register(atexit_rmtree, store)
    atexit.register(atexit_rmtree, chunk_store)
    path = "g1"
    g = open_group(store, path=path, chunk_store=chunk_store, zarr_version=zarr_version)
    assert g._store._store_version == g._chunk_store._store_version == zarr_version



import re
from unittest import mock

import numpy as np
import pytest

from zarr.core import Array
from zarr.util import (
    ConstantMap,
    all_equal,
    flatten,
    guess_chunks,
    human_readable_size,
    info_html_report,
    info_text_report,
    is_total_slice,
    json_dumps,
    normalize_chunks,
    normalize_dimension_separator,
    normalize_fill_value,
    normalize_order,
    normalize_resize_args,
    normalize_shape,
    retry_call,
    tree_array_icon,
    tree_group_icon,
    tree_get_icon,
    tree_widget,
)


def test_normalize_dimension_separator():
    assert None is normalize_dimension_separator(None)
    assert "/" == normalize_dimension_separator("/")
    assert "." == normalize_dimension_separator(".")
    with pytest.raises(ValueError):
        normalize_dimension_separator("X")


def test_normalize_shape():
    assert (100,) == normalize_shape((100,))
    assert (100,) == normalize_shape([100])
    assert (100,) == normalize_shape(100)
    with pytest.raises(TypeError):
        normalize_shape(None)
    with pytest.raises(ValueError):
        with pytest.warns(UserWarning):
            normalize_shape("foo")


def test_normalize_chunks():
    assert (10,) == normalize_chunks((10,), (100,), 1)
    assert (10,) == normalize_chunks([10], (100,), 1)
    assert (10,) == normalize_chunks(10, (100,), 1)
    assert (10, 10) == normalize_chunks((10, 10), (100, 10), 1)
    assert (10, 10) == normalize_chunks(10, (100, 10), 1)
    assert (10, 10) == normalize_chunks((10, None), (100, 10), 1)
    assert (30, 30, 30) == normalize_chunks(30, (100, 20, 10), 1)
    assert (30, 20, 10) == normalize_chunks((30,), (100, 20, 10), 1)
    assert (30, 20, 10) == normalize_chunks((30, None), (100, 20, 10), 1)
    assert (30, 20, 10) == normalize_chunks((30, None, None), (100, 20, 10), 1)
    assert (30, 20, 10) == normalize_chunks((30, 20, None), (100, 20, 10), 1)
    assert (30, 20, 10) == normalize_chunks((30, 20, 10), (100, 20, 10), 1)
    with pytest.raises(ValueError):
        normalize_chunks("foo", (100,), 1)
    with pytest.raises(ValueError):
        normalize_chunks((100, 10), (100,), 1)

    # test auto-chunking
    assert (100,) == normalize_chunks(None, (100,), 1)
    assert (100,) == normalize_chunks(-1, (100,), 1)
    assert (30, 20, 10) == normalize_chunks((30, -1, None), (100, 20, 10), 1)


def test_is_total_slice():
    # 1D
    assert is_total_slice(Ellipsis, (100,))
    assert is_total_slice(slice(None), (100,))
    assert is_total_slice(slice(0, 100), (100,))
    assert not is_total_slice(slice(0, 50), (100,))
    assert not is_total_slice(slice(0, 100, 2), (100,))

    # 2D
    assert is_total_slice(Ellipsis, (100, 100))
    assert is_total_slice(slice(None), (100, 100))
    assert is_total_slice((slice(None), slice(None)), (100, 100))
    assert is_total_slice((slice(0, 100), slice(0, 100)), (100, 100))
    assert not is_total_slice((slice(0, 100), slice(0, 50)), (100, 100))
    assert not is_total_slice((slice(0, 50), slice(0, 100)), (100, 100))
    assert not is_total_slice((slice(0, 50), slice(0, 50)), (100, 100))
    assert not is_total_slice((slice(0, 100, 2), slice(0, 100)), (100, 100))

    # size-1 dimension edge-case
    # https://github.com/zarr-developers/zarr-python/issues/1730
    assert is_total_slice((slice(0, 1),), (1,))
    # this is an equivalent selection (without a slice)
    assert is_total_slice((0,), (1,))
    # same for multidimensional selection
    assert is_total_slice((slice(0, 1), slice(0, 10)), (1, 10))
    assert is_total_slice((0, slice(0, 10)), (1, 10))

    with pytest.raises(TypeError):
        is_total_slice("foo", (100,))


def test_normalize_resize_args():
    # 1D
    assert (200,) == normalize_resize_args((100,), 200)
    assert (200,) == normalize_resize_args((100,), (200,))

    # 2D
    assert (200, 100) == normalize_resize_args((100, 100), (200, 100))
    assert (200, 100) == normalize_resize_args((100, 100), (200, None))
    assert (200, 100) == normalize_resize_args((100, 100), 200, 100)
    assert (200, 100) == normalize_resize_args((100, 100), 200, None)

    with pytest.raises(ValueError):
        normalize_resize_args((100,), (200, 100))


def test_human_readable_size():
    assert "100" == human_readable_size(100)
    assert "1.0K" == human_readable_size(2**10)
    assert "1.0M" == human_readable_size(2**20)
    assert "1.0G" == human_readable_size(2**30)
    assert "1.0T" == human_readable_size(2**40)
    assert "1.0P" == human_readable_size(2**50)


def test_normalize_order():
    assert "F" == normalize_order("F")
    assert "C" == normalize_order("C")
    assert "F" == normalize_order("f")
    assert "C" == normalize_order("c")
    with pytest.raises(ValueError):
        normalize_order("foo")


def test_normalize_fill_value():
    assert b"" == normalize_fill_value(0, dtype=np.dtype("S1"))
    structured_dtype = np.dtype([("foo", "S3"), ("bar", "i4"), ("baz", "f8")])
    expect = np.array((b"", 0, 0.0), dtype=structured_dtype)[()]
    assert expect == normalize_fill_value(0, dtype=structured_dtype)
    assert expect == normalize_fill_value(expect, dtype=structured_dtype)
    assert "" == normalize_fill_value(0, dtype=np.dtype("U1"))


def test_guess_chunks():
    shapes = (
        (100,),
        (100, 100),
        (1000000,),
        (1000000000,),
        (10000000000000000000000,),
        (10000, 10000),
        (10000000, 1000),
        (1000, 10000000),
        (10000000, 1000, 2),
        (1000, 10000000, 2),
        (10000, 10000, 10000),
        (100000, 100000, 100000),
        (1000000000, 1000000000, 1000000000),
        (0,),
        (0, 0),
        (10, 0),
        (0, 10),
        (1, 2, 0, 4, 5),
    )
    for shape in shapes:
        chunks = guess_chunks(shape, 1)
        assert isinstance(chunks, tuple)
        assert len(chunks) == len(shape)
        # doesn't make any sense to allow chunks to have zero length dimension
        assert all(0 < c <= max(s, 1) for c, s in zip(chunks, shape))

    # ludicrous itemsize
    chunks = guess_chunks((1000000,), 40000000000)
    assert isinstance(chunks, tuple)
    assert (1,) == chunks


def test_info_text_report():
    items = [("foo", "bar"), ("baz", "qux")]
    expect = "foo : bar\nbaz : qux\n"
    assert expect == info_text_report(items)


def test_info_html_report():
    items = [("foo", "bar"), ("baz", "qux")]
    actual = info_html_report(items)
    assert "<table" == actual[:6]
    assert "</table>" == actual[-8:]


def test_tree_get_icon():
    assert tree_get_icon("Array") == tree_array_icon
    assert tree_get_icon("Group") == tree_group_icon
    with pytest.raises(ValueError):
        tree_get_icon("Baz")


@mock.patch.dict("sys.modules", {"ipytree": None})
def test_tree_widget_missing_ipytree():
    pattern = (
        "Run `pip install zarr[jupyter]` or `conda install ipytree`"
        "to get the required ipytree dependency for displaying the tree "
        "widget. If using jupyterlab<3, you also need to run "
        "`jupyter labextension install ipytree`"
    )
    with pytest.raises(ImportError, match=re.escape(pattern)):
        tree_widget(None, None, None)


def test_retry_call():
    class Fixture:
        def __init__(self, pass_on=1):
            self.c = 0
            self.pass_on = pass_on

        def __call__(self):
            self.c += 1
            if self.c != self.pass_on:
                raise PermissionError()

    for x in range(1, 11):
        # Any number of failures less than 10 will be accepted.
        fixture = Fixture(pass_on=x)
        retry_call(fixture, exceptions=(PermissionError,), wait=0)
        assert fixture.c == x

    def fail(x):
        # Failures after 10 will cause an error to be raised.
        retry_call(Fixture(pass_on=x), exceptions=(Exception,), wait=0)

    for x in range(11, 15):
        pytest.raises(PermissionError, fail, x)


def test_flatten():
    assert list(
        flatten(
            [
                "0",
                [
                    "1",
                    [
                        "2",
                        [
                            "3",
                            [
                                4,
                            ],
                        ],
                    ],
                ],
            ]
        )
    ) == ["0", "1", "2", "3", 4]
    assert list(flatten("foo")) == ["f", "o", "o"]
    assert list(flatten(["foo"])) == ["foo"]


def test_all_equal():
    assert all_equal(0, np.zeros((10, 10, 10)))
    assert not all_equal(1, np.zeros((10, 10, 10)))

    assert all_equal(1, np.ones((10, 10, 10)))
    assert not all_equal(1, 1 + np.ones((10, 10, 10)))

    assert all_equal(np.nan, np.array([np.nan, np.nan]))
    assert not all_equal(np.nan, np.array([np.nan, 1.0]))

    assert all_equal({"a": -1}, np.array([{"a": -1}, {"a": -1}], dtype="object"))
    assert not all_equal({"a": -1}, np.array([{"a": -1}, {"a": 2}], dtype="object"))

    assert all_equal(np.timedelta64(999, "D"), np.array([999, 999], dtype="timedelta64[D]"))
    assert not all_equal(np.timedelta64(999, "D"), np.array([999, 998], dtype="timedelta64[D]"))

    # all_equal(None, *) always returns False
    assert not all_equal(None, np.array([None, None]))
    assert not all_equal(None, np.array([None, 10]))


def test_json_dumps_numpy_dtype():
    assert json_dumps(np.int64(0)) == json_dumps(0)
    assert json_dumps(np.float32(0)) == json_dumps(float(0))
    # Check that we raise the error of the superclass for unsupported object
    with pytest.raises(TypeError):
        json_dumps(Array)


def test_constant_map():
    val = object()
    m = ConstantMap(keys=[1, 2], constant=val)
    assert len(m) == 2
    assert m[1] is val
    assert m[2] is val
    assert 1 in m
    assert 0 not in m
    with pytest.raises(KeyError):
        m[0]
    assert repr(m) == repr({1: val, 2: val})



# flake8: noqa
from zarr.codecs import *
from zarr.convenience import (
    consolidate_metadata,
    copy,
    copy_all,
    copy_store,
    load,
    open,
    open_consolidated,
    save,
    save_array,
    save_group,
    tree,
)
from zarr.core import Array
from zarr.creation import (
    array,
    create,
    empty,
    empty_like,
    full,
    full_like,
    ones,
    ones_like,
    open_array,
    open_like,
    zeros,
    zeros_like,
)
from zarr.errors import CopyError, MetadataError
from zarr.hierarchy import Group, group, open_group
from zarr.n5 import N5Store, N5FSStore
from zarr._storage.store import v3_api_available
from zarr.storage import (
    ABSStore,
    DBMStore,
    DictStore,
    DirectoryStore,
    KVStore,
    LMDBStore,
    LRUStoreCache,
    MemoryStore,
    MongoDBStore,
    NestedDirectoryStore,
    RedisStore,
    SQLiteStore,
    TempStore,
    ZipStore,
)
from zarr.sync import ProcessSynchronizer, ThreadSynchronizer
from zarr.version import version as __version__

# in case setuptools scm screw up and find version to be 0.0.0
assert not __version__.startswith("0.0.0")

if v3_api_available:
    from zarr._storage.v3 import (
        ABSStoreV3,
        DBMStoreV3,
        KVStoreV3,
        DirectoryStoreV3,
        LMDBStoreV3,
        LRUStoreCacheV3,
        MemoryStoreV3,
        MongoDBStoreV3,
        RedisStoreV3,
        SQLiteStoreV3,
        ZipStoreV3,
    )



import sys
import timeit

import numpy as np

import line_profiler
import zarr
from zarr import blosc

if __name__ == "__main__":
    sys.path.insert(0, "..")

    # setup
    a = np.random.normal(2000, 1000, size=200000000).astype("u2")
    z = zarr.empty_like(
        a,
        chunks=1000000,
        compression="blosc",
        compression_opts=dict(cname="lz4", clevel=5, shuffle=2),
    )
    print(z)

    print("*" * 79)

    # time
    t = timeit.repeat("z[:] = a", repeat=10, number=1, globals=globals())
    print(t)
    print(min(t))
    print(z)

    # profile
    profile = line_profiler.LineProfiler(blosc.compress)
    profile.run("z[:] = a")
    profile.print_stats()

    print("*" * 79)

    # time
    t = timeit.repeat("z[:]", repeat=10, number=1, globals=globals())
    print(t)
    print(min(t))

    # profile
    profile = line_profiler.LineProfiler(blosc.decompress)
    profile.run("z[:]")
    profile.print_stats()



import pytest

from zarr.n5 import N5ChunkWrapper, N5FSStore
from zarr.creation import create
from zarr.storage import atexit_rmtree
from numcodecs import GZip
import numpy as np
from typing import Tuple
import json
import atexit

from zarr.tests.util import have_fsspec


def test_make_n5_chunk_wrapper():
    dtype = "uint8"
    chunk_shape = (10,)
    codec = GZip()
    # ValueError when specifying both compressor and compressor_config
    with pytest.raises(ValueError):
        N5ChunkWrapper(
            dtype, chunk_shape=chunk_shape, compressor_config=codec.get_config(), compressor=codec
        )

    wrapper_a = N5ChunkWrapper(dtype, chunk_shape=chunk_shape, compressor_config=codec.get_config())
    wrapper_b = N5ChunkWrapper(dtype, chunk_shape=chunk_shape, compressor=codec)
    assert wrapper_a == wrapper_b


@pytest.mark.parametrize("chunk_shape", ((2,), (4, 4), (8, 8, 8)))
def test_partial_chunk_decode(chunk_shape: Tuple[int, ...]):
    # Test that the N5Chunk wrapper can handle fractional chunks that
    # may be generated by other N5 implementations
    dtype = "uint8"
    codec = GZip()
    codec_wrapped = N5ChunkWrapper(dtype, chunk_shape=chunk_shape, compressor=codec)
    subslices = tuple(slice(0, cs // 2) for cs in chunk_shape)
    chunk = np.zeros(chunk_shape, dtype=dtype)
    chunk[subslices] = 1
    subchunk = np.ascontiguousarray(chunk[subslices])
    assert np.array_equal(codec_wrapped.decode(codec_wrapped.encode(subchunk)), chunk)


@pytest.mark.skipif(have_fsspec is False, reason="needs fsspec")
def test_dtype_decode():
    path = "data/array.n5"
    atexit_rmtree(path)
    atexit.register(atexit_rmtree, path)
    n5_store = N5FSStore(path)
    create(100, store=n5_store)
    dtype_n5 = json.loads(n5_store[".zarray"])["dtype"]
    dtype_zarr = json.loads(create(100).store[".zarray"])["dtype"]
    assert dtype_n5 == dtype_zarr



from typing import Optional
import numpy as np
import pytest

from numcodecs.abc import Codec
from numcodecs.compat import ensure_contiguous_ndarray_like
from numcodecs.registry import get_codec, register_codec

import zarr.codecs
from zarr.core import Array
from zarr.creation import array, empty, full, ones, open_array, zeros
from zarr.hierarchy import open_group, group
from zarr.storage import DirectoryStore, MemoryStore, Store, ZipStore


class CuPyCPUCompressor(Codec):  # pragma: no cover
    """CPU compressor for CuPy arrays

    This compressor converts CuPy arrays host memory before compressing
    the arrays using `compressor`.

    Parameters
    ----------
    compressor : numcodecs.abc.Codec
        The codec to use for compression and decompression.
    """

    codec_id = "cupy_cpu_compressor"

    def __init__(self, compressor: Codec = None):
        self.compressor = compressor

    def encode(self, buf):
        import cupy

        buf = cupy.asnumpy(ensure_contiguous_ndarray_like(buf))
        if self.compressor:
            buf = self.compressor.encode(buf)
        return buf

    def decode(self, chunk, out=None):
        import cupy

        if self.compressor:
            cpu_out = None if out is None else cupy.asnumpy(out)
            chunk = self.compressor.decode(chunk, cpu_out)

        chunk = cupy.asarray(ensure_contiguous_ndarray_like(chunk))
        if out is not None:
            cupy.copyto(out, chunk.view(dtype=out.dtype), casting="no")
            chunk = out
        return chunk

    def get_config(self):
        cc_config = self.compressor.get_config() if self.compressor else None
        return {
            "id": self.codec_id,
            "compressor_config": cc_config,
        }

    @classmethod
    def from_config(cls, config):
        cc_config = config.get("compressor_config", None)
        compressor = get_codec(cc_config) if cc_config else None
        return cls(compressor=compressor)


register_codec(CuPyCPUCompressor)


class MyArray(np.ndarray):
    """Dummy array class to test the `meta_array` argument

    Useful when CuPy isn't available.

    This class also makes some of the functions from the numpy
    module available.
    """

    testing = np.testing

    @classmethod
    def arange(cls, size):
        ret = cls(shape=(size,), dtype="int64")
        ret[:] = range(size)
        return ret

    @classmethod
    def empty(cls, shape):
        return cls(shape=shape)


def init_compressor(compressor) -> CuPyCPUCompressor:
    if compressor:
        compressor = getattr(zarr.codecs, compressor)()
    return CuPyCPUCompressor(compressor)


def init_store(tmp_path, store_type) -> Optional[Store]:
    if store_type is DirectoryStore:
        return store_type(str(tmp_path / "store"))
    if store_type is MemoryStore:
        return MemoryStore()
    return None


def ensure_module(module):
    if isinstance(module, str):
        return pytest.importorskip(module)
    return module


param_module_and_compressor = [
    (MyArray, None),
    ("cupy", init_compressor(None)),
    ("cupy", init_compressor("Zlib")),
    ("cupy", init_compressor("Blosc")),
]


@pytest.mark.parametrize("module, compressor", param_module_and_compressor)
@pytest.mark.parametrize("store_type", [None, DirectoryStore, MemoryStore, ZipStore])
def test_array(tmp_path, module, compressor, store_type):
    xp = ensure_module(module)

    store = init_store(tmp_path / "from_cupy_array", store_type)
    a = xp.arange(100)
    z = array(a, chunks=10, compressor=compressor, store=store, meta_array=xp.empty(()))
    assert a.shape == z.shape
    assert a.dtype == z.dtype
    assert isinstance(a, type(z[:]))
    assert isinstance(z.meta_array, type(xp.empty(())))
    xp.testing.assert_array_equal(a, z[:])

    # with array-like
    store = init_store(tmp_path / "from_list", store_type)
    a = list(range(100))
    z = array(a, chunks=10, compressor=compressor, store=store, meta_array=xp.empty(()))
    assert (100,) == z.shape
    assert np.asarray(a).dtype == z.dtype
    xp.testing.assert_array_equal(a, z[:])

    # with another zarr array
    store = init_store(tmp_path / "from_another_store", store_type)
    z2 = array(z, compressor=compressor, store=store, meta_array=xp.empty(()))
    assert z.shape == z2.shape
    assert z.chunks == z2.chunks
    assert z.dtype == z2.dtype
    xp.testing.assert_array_equal(z[:], z2[:])

    store = init_store(tmp_path / "open_array", store_type)
    a = xp.arange(100)
    z = open_array(
        store,
        shape=a.shape,
        dtype=a.dtype,
        chunks=10,
        compressor=compressor,
        meta_array=xp.empty(()),
    )
    z[:] = a
    assert a.shape == z.shape
    assert a.dtype == z.dtype
    assert isinstance(a, type(z[:]))
    assert isinstance(z.meta_array, type(xp.empty(())))
    xp.testing.assert_array_equal(a, z[:])


@pytest.mark.parametrize("module, compressor", param_module_and_compressor)
def test_empty(module, compressor):
    xp = ensure_module(module)
    z = empty(
        100,
        chunks=10,
        compressor=compressor,
        meta_array=xp.empty(()),
    )
    assert (100,) == z.shape
    assert (10,) == z.chunks


@pytest.mark.parametrize("module, compressor", param_module_and_compressor)
def test_zeros(module, compressor):
    xp = ensure_module(module)
    z = zeros(
        100,
        chunks=10,
        compressor=compressor,
        meta_array=xp.empty(()),
    )
    assert (100,) == z.shape
    assert (10,) == z.chunks
    xp.testing.assert_array_equal(np.zeros(100), z[:])


@pytest.mark.parametrize("module, compressor", param_module_and_compressor)
def test_ones(module, compressor):
    xp = ensure_module(module)
    z = ones(
        100,
        chunks=10,
        compressor=compressor,
        meta_array=xp.empty(()),
    )
    assert (100,) == z.shape
    assert (10,) == z.chunks
    xp.testing.assert_array_equal(np.ones(100), z[:])


@pytest.mark.parametrize("module, compressor", param_module_and_compressor)
def test_full(module, compressor):
    xp = ensure_module(module)
    z = full(
        100,
        chunks=10,
        fill_value=42,
        dtype="i4",
        compressor=compressor,
        meta_array=xp.empty(()),
    )
    assert (100,) == z.shape
    assert (10,) == z.chunks
    xp.testing.assert_array_equal(np.full(100, fill_value=42, dtype="i4"), z[:])

    # nan
    z = full(
        100,
        chunks=10,
        fill_value=np.nan,
        dtype="f8",
        compressor=compressor,
        meta_array=xp.empty(()),
    )
    assert np.all(np.isnan(z[:]))


@pytest.mark.parametrize("group_create_function", [group, open_group])
@pytest.mark.parametrize("module, compressor", param_module_and_compressor)
@pytest.mark.parametrize("store_type", [None, DirectoryStore, MemoryStore, ZipStore])
def test_group(tmp_path, group_create_function, module, compressor, store_type):
    xp = ensure_module(module)
    store = init_store(tmp_path, store_type)
    g = group_create_function(store, meta_array=xp.empty(()))
    g.ones("data", shape=(10, 11), dtype=int, compressor=compressor)
    a = g["data"]
    assert a.shape == (10, 11)
    assert a.dtype == int
    assert isinstance(a, Array)
    assert isinstance(a[:], type(xp.empty(())))
    assert (a[:] == 1).all()
    assert isinstance(g.meta_array, type(xp.empty(())))



import atexit
import tempfile
import unittest
from numbers import Integral

import numpy as np
import pytest
from numcodecs import Adler32, Zlib
from numpy.testing import assert_array_equal

import zarr
from zarr.convenience import (
    consolidate_metadata,
    copy,
    copy_store,
    load,
    open,
    open_consolidated,
    save,
    save_group,
    save_array,
    copy_all,
)
from zarr.core import Array
from zarr.errors import CopyError
from zarr.hierarchy import Group, group
from zarr.storage import (
    ConsolidatedMetadataStore,
    FSStore,
    KVStore,
    MemoryStore,
    atexit_rmtree,
    data_root,
    meta_root,
    getsize,
)
from zarr._storage.store import v3_api_available
from zarr._storage.v3 import (
    ConsolidatedMetadataStoreV3,
    DirectoryStoreV3,
    FSStoreV3,
    KVStoreV3,
    MemoryStoreV3,
    SQLiteStoreV3,
)
from zarr.tests.util import have_fsspec

_VERSIONS = (2, 3) if v3_api_available else (2,)


def _init_creation_kwargs(zarr_version):
    kwargs = {"zarr_version": zarr_version}
    if zarr_version == 3:
        kwargs["path"] = "dataset"
    return kwargs


@pytest.mark.parametrize("zarr_version", _VERSIONS)
def test_open_array(path_type, zarr_version):
    store = tempfile.mkdtemp()
    atexit.register(atexit_rmtree, store)
    store = path_type(store)
    kwargs = _init_creation_kwargs(zarr_version)

    # open array, create if doesn't exist
    z = open(store, mode="a", shape=100, **kwargs)
    assert isinstance(z, Array)
    assert z.shape == (100,)

    # open array, overwrite
    z = open(store, mode="w", shape=200, **kwargs)
    assert isinstance(z, Array)
    assert z.shape == (200,)

    # open array, read-only
    z = open(store, mode="r", **kwargs)
    assert isinstance(z, Array)
    assert z.shape == (200,)
    assert z.read_only

    # path not found
    with pytest.raises(ValueError):
        open("doesnotexist", mode="r")


@pytest.mark.parametrize("zarr_version", _VERSIONS)
def test_open_group(path_type, zarr_version):
    store = tempfile.mkdtemp()
    atexit.register(atexit_rmtree, store)
    store = path_type(store)
    kwargs = _init_creation_kwargs(zarr_version)

    # open group, create if doesn't exist
    g = open(store, mode="a", **kwargs)
    g.create_group("foo")
    assert isinstance(g, Group)
    assert "foo" in g

    # open group, overwrite
    g = open(store, mode="w", **kwargs)
    assert isinstance(g, Group)
    assert "foo" not in g

    # open group, read-only
    g = open(store, mode="r", **kwargs)
    assert isinstance(g, Group)
    assert g.read_only


@pytest.mark.parametrize("zarr_version", _VERSIONS)
def test_save_errors(zarr_version):
    with pytest.raises(ValueError):
        # no arrays provided
        save_group("data/group.zarr", zarr_version=zarr_version)
    with pytest.raises(TypeError):
        # no array provided
        save_array("data/group.zarr", zarr_version=zarr_version)
    with pytest.raises(ValueError):
        # no arrays provided
        save("data/group.zarr", zarr_version=zarr_version)


@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
def test_zarr_v3_save_multiple_unnamed():
    x = np.ones(8)
    y = np.zeros(8)
    store = KVStoreV3(dict())
    # no path provided
    save_group(store, x, y, path="dataset", zarr_version=3)
    # names become arr_{i} for unnamed *args
    assert data_root + "dataset/arr_0/c0" in store
    assert data_root + "dataset/arr_1/c0" in store
    assert meta_root + "dataset/arr_0.array.json" in store
    assert meta_root + "dataset/arr_1.array.json" in store


@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
def test_zarr_v3_save_errors():
    x = np.ones(8)
    with pytest.raises(ValueError):
        # no path provided
        save_group("data/group.zr3", x, zarr_version=3)
    with pytest.raises(ValueError):
        # no path provided
        save_array("data/group.zr3", x, zarr_version=3)
    with pytest.raises(ValueError):
        # no path provided
        save("data/group.zr3", x, zarr_version=3)


@pytest.mark.parametrize("zarr_version", _VERSIONS)
def test_lazy_loader(zarr_version):
    foo = np.arange(100)
    bar = np.arange(100, 0, -1)
    store = "data/group.zarr" if zarr_version == 2 else "data/group.zr3"
    kwargs = _init_creation_kwargs(zarr_version)
    save(store, foo=foo, bar=bar, **kwargs)
    loader = load(store, **kwargs)
    assert "foo" in loader
    assert "bar" in loader
    assert "baz" not in loader
    assert len(loader) == 2
    assert sorted(loader) == ["bar", "foo"]
    assert_array_equal(foo, loader["foo"])
    assert_array_equal(bar, loader["bar"])
    assert "LazyLoader: " in repr(loader)


@pytest.mark.parametrize("zarr_version", _VERSIONS)
def test_load_array(zarr_version):
    foo = np.arange(100)
    bar = np.arange(100, 0, -1)
    store = "data/group.zarr" if zarr_version == 2 else "data/group.zr3"
    kwargs = _init_creation_kwargs(zarr_version)
    save(store, foo=foo, bar=bar, **kwargs)

    # can also load arrays directly into a numpy array
    for array_name in ["foo", "bar"]:
        array_path = "dataset/" + array_name if zarr_version == 3 else array_name
        array = load(store, path=array_path, zarr_version=zarr_version)
        assert isinstance(array, np.ndarray)
        if array_name == "foo":
            assert_array_equal(foo, array)
        else:
            assert_array_equal(bar, array)


@pytest.mark.parametrize("zarr_version", _VERSIONS)
def test_tree(zarr_version):
    kwargs = _init_creation_kwargs(zarr_version)
    g1 = zarr.group(**kwargs)
    g1.create_group("foo")
    g3 = g1.create_group("bar")
    g3.create_group("baz")
    g5 = g3.create_group("qux")
    g5.create_dataset("baz", shape=100, chunks=10)
    assert repr(zarr.tree(g1)) == repr(g1.tree())
    assert str(zarr.tree(g1)) == str(g1.tree())


@pytest.mark.parametrize("zarr_version", _VERSIONS)
@pytest.mark.parametrize("stores_from_path", [False, True])
@pytest.mark.parametrize(
    "with_chunk_store,listable",
    [(False, True), (True, True), (False, False)],
    ids=["default-listable", "with_chunk_store-listable", "default-unlistable"],
)
def test_consolidate_metadata(
    with_chunk_store, zarr_version, listable, monkeypatch, stores_from_path
):
    # setup initial data
    if stores_from_path:
        store = tempfile.mkdtemp()
        atexit.register(atexit_rmtree, store)
        if with_chunk_store:
            chunk_store = tempfile.mkdtemp()
            atexit.register(atexit_rmtree, chunk_store)
        else:
            chunk_store = None
        version_kwarg = {"zarr_version": zarr_version}
    else:
        if zarr_version == 2:
            store = MemoryStore()
            chunk_store = MemoryStore() if with_chunk_store else None
        elif zarr_version == 3:
            store = MemoryStoreV3()
            chunk_store = MemoryStoreV3() if with_chunk_store else None
        version_kwarg = {}
    path = "dataset" if zarr_version == 3 else None
    z = group(store, chunk_store=chunk_store, path=path, **version_kwarg)

    # Reload the actual store implementation in case str
    store_to_copy = z.store

    z.create_group("g1")
    g2 = z.create_group("g2")
    g2.attrs["hello"] = "world"
    arr = g2.create_dataset("arr", shape=(20, 20), chunks=(5, 5), dtype="f8")
    assert 16 == arr.nchunks
    assert 0 == arr.nchunks_initialized
    arr.attrs["data"] = 1
    arr[:] = 1.0
    assert 16 == arr.nchunks_initialized

    if stores_from_path:
        # get the actual store class for use with consolidate_metadata
        store_class = z._store
    else:
        store_class = store

    if zarr_version == 3:
        # error on v3 if path not provided
        with pytest.raises(ValueError):
            consolidate_metadata(store_class, path=None)

        with pytest.raises(ValueError):
            consolidate_metadata(store_class, path="")

    # perform consolidation
    out = consolidate_metadata(store_class, path=path)
    assert isinstance(out, Group)
    assert ["g1", "g2"] == list(out)
    if not stores_from_path:
        if zarr_version == 2:
            assert isinstance(out._store, ConsolidatedMetadataStore)
            assert ".zmetadata" in store
            meta_keys = [
                ".zgroup",
                "g1/.zgroup",
                "g2/.zgroup",
                "g2/.zattrs",
                "g2/arr/.zarray",
                "g2/arr/.zattrs",
            ]
        else:
            assert isinstance(out._store, ConsolidatedMetadataStoreV3)
            assert "meta/root/consolidated/.zmetadata" in store
            meta_keys = [
                "zarr.json",
                meta_root + "dataset.group.json",
                meta_root + "dataset/g1.group.json",
                meta_root + "dataset/g2.group.json",
                meta_root + "dataset/g2/arr.array.json",
                "meta/root/consolidated.group.json",
            ]
        for key in meta_keys:
            del store[key]

    # https://github.com/zarr-developers/zarr-python/issues/993
    # Make sure we can still open consolidated on an unlistable store:
    if not listable:
        fs_memory = pytest.importorskip("fsspec.implementations.memory")
        monkeypatch.setattr(fs_memory.MemoryFileSystem, "isdir", lambda x, y: False)
        monkeypatch.delattr(fs_memory.MemoryFileSystem, "ls")
        fs = fs_memory.MemoryFileSystem()
        if zarr_version == 2:
            store_to_open = FSStore("", fs=fs)
        else:
            store_to_open = FSStoreV3("", fs=fs)

        # copy original store to new unlistable store
        store_to_open.update(store_to_copy)

    else:
        store_to_open = store

    # open consolidated
    z2 = open_consolidated(store_to_open, chunk_store=chunk_store, path=path, **version_kwarg)
    assert ["g1", "g2"] == list(z2)
    assert "world" == z2.g2.attrs["hello"]
    assert 1 == z2.g2.arr.attrs["data"]
    assert (z2.g2.arr[:] == 1.0).all()
    assert 16 == z2.g2.arr.nchunks
    if listable:
        assert 16 == z2.g2.arr.nchunks_initialized
    else:
        with pytest.raises(NotImplementedError):
            _ = z2.g2.arr.nchunks_initialized

    if stores_from_path:
        # path string is note a BaseStore subclass so cannot be used to
        # initialize a ConsolidatedMetadataStore.
        if zarr_version == 2:
            with pytest.raises(ValueError):
                cmd = ConsolidatedMetadataStore(store)
        elif zarr_version == 3:
            with pytest.raises(ValueError):
                cmd = ConsolidatedMetadataStoreV3(store)
    else:
        # tests del/write on the store
        if zarr_version == 2:
            cmd = ConsolidatedMetadataStore(store)
            with pytest.raises(PermissionError):
                del cmd[".zgroup"]
            with pytest.raises(PermissionError):
                cmd[".zgroup"] = None
        else:
            cmd = ConsolidatedMetadataStoreV3(store)
            with pytest.raises(PermissionError):
                del cmd[meta_root + "dataset.group.json"]
            with pytest.raises(PermissionError):
                cmd[meta_root + "dataset.group.json"] = None

        # test getsize on the store
        assert isinstance(getsize(cmd), Integral)

    # test new metadata are not writeable
    with pytest.raises(PermissionError):
        z2.create_group("g3")
    with pytest.raises(PermissionError):
        z2.create_dataset("spam", shape=42, chunks=7, dtype="i4")
    with pytest.raises(PermissionError):
        del z2["g2"]

    # test consolidated metadata are not writeable
    with pytest.raises(PermissionError):
        z2.g2.attrs["hello"] = "universe"
    with pytest.raises(PermissionError):
        z2.g2.arr.attrs["foo"] = "bar"

    # test the data are writeable
    z2.g2.arr[:] = 2
    assert (z2.g2.arr[:] == 2).all()

    # test invalid modes
    with pytest.raises(ValueError):
        open_consolidated(store, chunk_store=chunk_store, mode="a", path=path)
    with pytest.raises(ValueError):
        open_consolidated(store, chunk_store=chunk_store, mode="w", path=path)
    with pytest.raises(ValueError):
        open_consolidated(store, chunk_store=chunk_store, mode="w-", path=path)

    # make sure keyword arguments are passed through without error
    open_consolidated(
        store,
        chunk_store=chunk_store,
        path=path,
        cache_attrs=True,
        synchronizer=None,
        **version_kwarg,
    )


@pytest.mark.parametrize(
    "options",
    (
        {"dimension_separator": "/"},
        {"dimension_separator": "."},
        {"dimension_separator": None},
    ),
)
def test_save_array_separator(tmpdir, options):
    data = np.arange(6).reshape((3, 2))
    url = tmpdir.join("test.zarr")
    save_array(url, data, **options)


class TestCopyStore(unittest.TestCase):
    _version = 2

    def setUp(self):
        source = dict()
        source["foo"] = b"xxx"
        source["bar/baz"] = b"yyy"
        source["bar/qux"] = b"zzz"
        self.source = source

    def _get_dest_store(self):
        return dict()

    def test_no_paths(self):
        source = self.source
        dest = self._get_dest_store()
        copy_store(source, dest)
        assert len(source) == len(dest)
        for key in source:
            assert source[key] == dest[key]

    def test_source_path(self):
        source = self.source
        # paths should be normalized
        for source_path in "bar", "bar/", "/bar", "/bar/":
            dest = self._get_dest_store()
            copy_store(source, dest, source_path=source_path)
            assert 2 == len(dest)
            for key in source:
                if key.startswith("bar/"):
                    dest_key = key.split("bar/")[1]
                    assert source[key] == dest[dest_key]
                else:
                    assert key not in dest

    def test_dest_path(self):
        source = self.source
        # paths should be normalized
        for dest_path in "new", "new/", "/new", "/new/":
            dest = self._get_dest_store()
            copy_store(source, dest, dest_path=dest_path)
            assert len(source) == len(dest)
            for key in source:
                if self._version == 3:
                    dest_key = key[:10] + "new/" + key[10:]
                else:
                    dest_key = "new/" + key
                assert source[key] == dest[dest_key]

    def test_source_dest_path(self):
        source = self.source
        # paths should be normalized
        for source_path in "bar", "bar/", "/bar", "/bar/":
            for dest_path in "new", "new/", "/new", "/new/":
                dest = self._get_dest_store()
                copy_store(source, dest, source_path=source_path, dest_path=dest_path)
                assert 2 == len(dest)
                for key in source:
                    if key.startswith("bar/"):
                        dest_key = "new/" + key.split("bar/")[1]
                        assert source[key] == dest[dest_key]
                    else:
                        assert key not in dest
                        assert ("new/" + key) not in dest

    def test_excludes_includes(self):
        source = self.source

        # single excludes
        dest = self._get_dest_store()
        excludes = "f.*"
        copy_store(source, dest, excludes=excludes)
        assert len(dest) == 2

        root = "" if self._version == 2 else meta_root
        assert root + "foo" not in dest

        # multiple excludes
        dest = self._get_dest_store()
        excludes = "b.z", ".*x"
        copy_store(source, dest, excludes=excludes)
        assert len(dest) == 1
        assert root + "foo" in dest
        assert root + "bar/baz" not in dest
        assert root + "bar/qux" not in dest

        # excludes and includes
        dest = self._get_dest_store()
        excludes = "b.*"
        includes = ".*x"
        copy_store(source, dest, excludes=excludes, includes=includes)
        assert len(dest) == 2
        assert root + "foo" in dest
        assert root + "bar/baz" not in dest
        assert root + "bar/qux" in dest

    def test_dry_run(self):
        source = self.source
        dest = self._get_dest_store()
        copy_store(source, dest, dry_run=True)
        assert 0 == len(dest)

    def test_if_exists(self):
        source = self.source
        dest = self._get_dest_store()
        root = "" if self._version == 2 else meta_root
        dest[root + "bar/baz"] = b"mmm"

        # default ('raise')
        with pytest.raises(CopyError):
            copy_store(source, dest)

        # explicit 'raise'
        with pytest.raises(CopyError):
            copy_store(source, dest, if_exists="raise")

        # skip
        copy_store(source, dest, if_exists="skip")
        assert 3 == len(dest)
        assert dest[root + "foo"] == b"xxx"
        assert dest[root + "bar/baz"] == b"mmm"
        assert dest[root + "bar/qux"] == b"zzz"

        # replace
        copy_store(source, dest, if_exists="replace")
        assert 3 == len(dest)
        assert dest[root + "foo"] == b"xxx"
        assert dest[root + "bar/baz"] == b"yyy"
        assert dest[root + "bar/qux"] == b"zzz"

        # invalid option
        with pytest.raises(ValueError):
            copy_store(source, dest, if_exists="foobar")


@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
class TestCopyStoreV3(TestCopyStore):
    _version = 3

    def setUp(self):
        source = KVStoreV3(dict())
        source["meta/root/foo"] = b"xxx"
        source["meta/root/bar/baz"] = b"yyy"
        source["meta/root/bar/qux"] = b"zzz"
        self.source = source

    def _get_dest_store(self):
        return KVStoreV3(dict())

    def test_mismatched_store_versions(self):
        # cannot copy between stores of mixed Zarr versions
        dest = KVStore(dict())
        with pytest.raises(ValueError):
            copy_store(self.source, dest)


def check_copied_array(original, copied, without_attrs=False, expect_props=None):
    # setup
    source_h5py = original.__module__.startswith("h5py.")
    dest_h5py = copied.__module__.startswith("h5py.")
    zarr_to_zarr = not (source_h5py or dest_h5py)
    h5py_to_h5py = source_h5py and dest_h5py
    zarr_to_h5py = not source_h5py and dest_h5py
    h5py_to_zarr = source_h5py and not dest_h5py
    if expect_props is None:
        expect_props = dict()
    else:
        expect_props = expect_props.copy()

    # common properties in zarr and h5py
    for p in "dtype", "shape", "chunks":
        expect_props.setdefault(p, getattr(original, p))

    # zarr-specific properties
    if zarr_to_zarr:
        for p in "compressor", "filters", "order", "fill_value":
            expect_props.setdefault(p, getattr(original, p))

    # h5py-specific properties
    if h5py_to_h5py:
        for p in (
            "maxshape",
            "compression",
            "compression_opts",
            "shuffle",
            "scaleoffset",
            "fletcher32",
            "fillvalue",
        ):
            expect_props.setdefault(p, getattr(original, p))

    # common properties with some name differences
    if h5py_to_zarr:
        expect_props.setdefault("fill_value", original.fillvalue)
    if zarr_to_h5py:
        expect_props.setdefault("fillvalue", original.fill_value)

    # compare properties
    for k, v in expect_props.items():
        assert v == getattr(copied, k)

    # compare data
    assert_array_equal(original[:], copied[:])

    # compare attrs
    if without_attrs:
        for k in original.attrs.keys():
            assert k not in copied.attrs
    else:
        if dest_h5py and "filters" in original.attrs:
            # special case in v3 (storing filters metadata under attributes)
            # we explicitly do not copy this info over to HDF5
            original_attrs = original.attrs.asdict().copy()
            original_attrs.pop("filters")
        else:
            original_attrs = original.attrs
        assert sorted(original_attrs.items()) == sorted(copied.attrs.items())


def check_copied_group(original, copied, without_attrs=False, expect_props=None, shallow=False):
    # setup
    if expect_props is None:
        expect_props = dict()
    else:
        expect_props = expect_props.copy()

    # compare children
    for k, v in original.items():
        if hasattr(v, "shape"):
            assert k in copied
            check_copied_array(v, copied[k], without_attrs=without_attrs, expect_props=expect_props)
        elif shallow:
            assert k not in copied
        else:
            assert k in copied
            check_copied_group(
                v,
                copied[k],
                without_attrs=without_attrs,
                shallow=shallow,
                expect_props=expect_props,
            )

    # compare attrs
    if without_attrs:
        for k in original.attrs.keys():
            assert k not in copied.attrs
    else:
        assert sorted(original.attrs.items()) == sorted(copied.attrs.items())


def test_copy_all():
    """
    https://github.com/zarr-developers/zarr-python/issues/269

    copy_all used to not copy attributes as `.keys()` does not return hidden `.zattrs`.

    """
    original_group = zarr.group(store=MemoryStore(), overwrite=True)
    original_group.attrs["info"] = "group attrs"
    original_subgroup = original_group.create_group("subgroup")
    original_subgroup.attrs["info"] = "sub attrs"

    destination_group = zarr.group(store=MemoryStore(), overwrite=True)

    # copy from memory to directory store
    copy_all(
        original_group,
        destination_group,
        dry_run=False,
    )

    assert "subgroup" in destination_group
    assert destination_group.attrs["info"] == "group attrs"
    assert destination_group.subgroup.attrs["info"] == "sub attrs"


@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
def test_copy_all_v3():
    """
    https://github.com/zarr-developers/zarr-python/issues/269

    copy_all used to not copy attributes as `.keys()`

    """
    original_group = zarr.group(store=MemoryStoreV3(), path="group1", overwrite=True)
    original_group.create_group("subgroup")

    destination_group = zarr.group(store=MemoryStoreV3(), path="group2", overwrite=True)

    # copy from memory to directory store
    copy_all(
        original_group,
        destination_group,
        dry_run=False,
    )
    assert "subgroup" in destination_group


class TestCopy:
    @pytest.fixture(params=[False, True], ids=["zarr", "hdf5"])
    def source(self, request, tmpdir):
        def prep_source(source):
            foo = source.create_group("foo")
            foo.attrs["experiment"] = "weird science"
            baz = foo.create_dataset("bar/baz", data=np.arange(100), chunks=(50,))
            baz.attrs["units"] = "metres"
            if request.param:
                extra_kws = dict(
                    compression="gzip",
                    compression_opts=3,
                    fillvalue=84,
                    shuffle=True,
                    fletcher32=True,
                )
            else:
                extra_kws = dict(compressor=Zlib(3), order="F", fill_value=42, filters=[Adler32()])
            source.create_dataset(
                "spam",
                data=np.arange(100, 200).reshape(20, 5),
                chunks=(10, 2),
                dtype="i2",
                **extra_kws,
            )
            return source

        if request.param:
            h5py = pytest.importorskip("h5py")
            fn = tmpdir.join("source.h5")
            with h5py.File(str(fn), mode="w") as h5f:
                yield prep_source(h5f)
        else:
            yield prep_source(group())

    @pytest.fixture(params=[False, True], ids=["zarr", "hdf5"])
    def dest(self, request, tmpdir):
        if request.param:
            h5py = pytest.importorskip("h5py")
            fn = tmpdir.join("dest.h5")
            with h5py.File(str(fn), mode="w") as h5f:
                yield h5f
        else:
            yield group()

    def test_copy_array(self, source, dest):
        # copy array with default options
        copy(source["foo/bar/baz"], dest)
        check_copied_array(source["foo/bar/baz"], dest["baz"])
        copy(source["spam"], dest)
        check_copied_array(source["spam"], dest["spam"])

    def test_copy_bad_dest(self, source, dest):
        # try to copy to an array, dest must be a group
        dest = dest.create_dataset("eggs", shape=(100,))
        with pytest.raises(ValueError):
            copy(source["foo/bar/baz"], dest)

    def test_copy_array_name(self, source, dest):
        # copy array with name
        copy(source["foo/bar/baz"], dest, name="qux")
        assert "baz" not in dest
        check_copied_array(source["foo/bar/baz"], dest["qux"])

    def test_copy_array_create_options(self, source, dest):
        dest_h5py = dest.__module__.startswith("h5py.")

        # copy array, provide creation options
        compressor = Zlib(9)
        create_kws = dict(chunks=(10,))
        if dest_h5py:
            create_kws.update(
                compression="gzip", compression_opts=9, shuffle=True, fletcher32=True, fillvalue=42
            )
        else:
            create_kws.update(compressor=compressor, fill_value=42, order="F", filters=[Adler32()])
        copy(source["foo/bar/baz"], dest, without_attrs=True, **create_kws)
        check_copied_array(
            source["foo/bar/baz"], dest["baz"], without_attrs=True, expect_props=create_kws
        )

    def test_copy_array_exists_array(self, source, dest):
        # copy array, dest array in the way
        dest.create_dataset("baz", shape=(10,))

        # raise
        with pytest.raises(CopyError):
            # should raise by default
            copy(source["foo/bar/baz"], dest)
        assert (10,) == dest["baz"].shape
        with pytest.raises(CopyError):
            copy(source["foo/bar/baz"], dest, if_exists="raise")
        assert (10,) == dest["baz"].shape

        # skip
        copy(source["foo/bar/baz"], dest, if_exists="skip")
        assert (10,) == dest["baz"].shape

        # replace
        copy(source["foo/bar/baz"], dest, if_exists="replace")
        check_copied_array(source["foo/bar/baz"], dest["baz"])

        # invalid option
        with pytest.raises(ValueError):
            copy(source["foo/bar/baz"], dest, if_exists="foobar")

    def test_copy_array_exists_group(self, source, dest):
        # copy array, dest group in the way
        dest.create_group("baz")

        # raise
        with pytest.raises(CopyError):
            copy(source["foo/bar/baz"], dest)
        assert not hasattr(dest["baz"], "shape")
        with pytest.raises(CopyError):
            copy(source["foo/bar/baz"], dest, if_exists="raise")
        assert not hasattr(dest["baz"], "shape")

        # skip
        copy(source["foo/bar/baz"], dest, if_exists="skip")
        assert not hasattr(dest["baz"], "shape")

        # replace
        copy(source["foo/bar/baz"], dest, if_exists="replace")
        check_copied_array(source["foo/bar/baz"], dest["baz"])

    def test_copy_array_skip_initialized(self, source, dest):
        dest_h5py = dest.__module__.startswith("h5py.")

        dest.create_dataset("baz", shape=(100,), chunks=(10,), dtype="i8")
        assert not np.all(source["foo/bar/baz"][:] == dest["baz"][:])

        if dest_h5py:
            with pytest.raises(ValueError):
                # not available with copy to h5py
                copy(source["foo/bar/baz"], dest, if_exists="skip_initialized")

        else:
            # copy array, dest array exists but not yet initialized
            copy(source["foo/bar/baz"], dest, if_exists="skip_initialized")
            check_copied_array(source["foo/bar/baz"], dest["baz"])

            # copy array, dest array exists and initialized, will be skipped
            dest["baz"][:] = np.arange(100, 200)
            copy(source["foo/bar/baz"], dest, if_exists="skip_initialized")
            assert_array_equal(np.arange(100, 200), dest["baz"][:])
            assert not np.all(source["foo/bar/baz"][:] == dest["baz"][:])

    def test_copy_group(self, source, dest):
        # copy group, default options
        copy(source["foo"], dest)
        check_copied_group(source["foo"], dest["foo"])

    def test_copy_group_no_name(self, source, dest):
        with pytest.raises(TypeError):
            # need a name if copy root
            copy(source, dest)

        copy(source, dest, name="root")
        check_copied_group(source, dest["root"])

    def test_copy_group_options(self, source, dest):
        # copy group, non-default options
        copy(source["foo"], dest, name="qux", without_attrs=True)
        assert "foo" not in dest
        check_copied_group(source["foo"], dest["qux"], without_attrs=True)

    def test_copy_group_shallow(self, source, dest):
        # copy group, shallow
        copy(source, dest, name="eggs", shallow=True)
        check_copied_group(source, dest["eggs"], shallow=True)

    def test_copy_group_exists_group(self, source, dest):
        # copy group, dest groups exist
        dest.create_group("foo/bar")
        copy(source["foo"], dest)
        check_copied_group(source["foo"], dest["foo"])

    def test_copy_group_exists_array(self, source, dest):
        # copy group, dest array in the way
        dest.create_dataset("foo/bar", shape=(10,))

        # raise
        with pytest.raises(CopyError):
            copy(source["foo"], dest)
        assert dest["foo/bar"].shape == (10,)
        with pytest.raises(CopyError):
            copy(source["foo"], dest, if_exists="raise")
        assert dest["foo/bar"].shape == (10,)

        # skip
        copy(source["foo"], dest, if_exists="skip")
        assert dest["foo/bar"].shape == (10,)

        # replace
        copy(source["foo"], dest, if_exists="replace")
        check_copied_group(source["foo"], dest["foo"])

    def test_copy_group_dry_run(self, source, dest):
        # dry run, empty destination
        n_copied, n_skipped, n_bytes_copied = copy(
            source["foo"], dest, dry_run=True, return_stats=True
        )
        assert 0 == len(dest)
        assert 3 == n_copied
        assert 0 == n_skipped
        assert 0 == n_bytes_copied

        # dry run, array exists in destination
        baz = np.arange(100, 200)
        dest.create_dataset("foo/bar/baz", data=baz)
        assert not np.all(source["foo/bar/baz"][:] == dest["foo/bar/baz"][:])
        assert 1 == len(dest)

        # raise
        with pytest.raises(CopyError):
            copy(source["foo"], dest, dry_run=True)
        assert 1 == len(dest)

        # skip
        n_copied, n_skipped, n_bytes_copied = copy(
            source["foo"], dest, dry_run=True, if_exists="skip", return_stats=True
        )
        assert 1 == len(dest)
        assert 2 == n_copied
        assert 1 == n_skipped
        assert 0 == n_bytes_copied
        assert_array_equal(baz, dest["foo/bar/baz"])

        # replace
        n_copied, n_skipped, n_bytes_copied = copy(
            source["foo"], dest, dry_run=True, if_exists="replace", return_stats=True
        )
        assert 1 == len(dest)
        assert 3 == n_copied
        assert 0 == n_skipped
        assert 0 == n_bytes_copied
        assert_array_equal(baz, dest["foo/bar/baz"])

    def test_logging(self, source, dest, tmpdir):
        # callable log
        copy(source["foo"], dest, dry_run=True, log=print)

        # file name
        fn = str(tmpdir.join("log_name"))
        copy(source["foo"], dest, dry_run=True, log=fn)

        # file
        with tmpdir.join("log_file").open(mode="w") as f:
            copy(source["foo"], dest, dry_run=True, log=f)

        # bad option
        with pytest.raises(TypeError):
            copy(source["foo"], dest, dry_run=True, log=True)


@pytest.mark.skipif(not v3_api_available, reason="V3 is disabled")
class TestCopyV3(TestCopy):
    @pytest.fixture(params=["zarr", "hdf5"])
    def source(self, request, tmpdir):
        def prep_source(source):
            foo = source.create_group("foo")
            foo.attrs["experiment"] = "weird science"
            baz = foo.create_dataset("bar/baz", data=np.arange(100), chunks=(50,))
            baz.attrs["units"] = "metres"
            if request.param == "hdf5":
                extra_kws = dict(
                    compression="gzip",
                    compression_opts=3,
                    fillvalue=84,
                    shuffle=True,
                    fletcher32=True,
                )
            else:
                extra_kws = dict(compressor=Zlib(3), order="F", fill_value=42, filters=[Adler32()])
            source.create_dataset(
                "spam",
                data=np.arange(100, 200).reshape(20, 5),
                chunks=(10, 2),
                dtype="i2",
                **extra_kws,
            )
            return source

        if request.param == "hdf5":
            h5py = pytest.importorskip("h5py")
            fn = tmpdir.join("source.h5")
            with h5py.File(str(fn), mode="w") as h5f:
                yield prep_source(h5f)
        elif request.param == "zarr":
            yield prep_source(group(path="group1", zarr_version=3))

    # Test with various destination StoreV3 types as TestCopyV3 covers rmdir
    destinations = ["hdf5", "zarr", "zarr_kvstore", "zarr_directorystore", "zarr_sqlitestore"]
    if have_fsspec:
        destinations += ["zarr_fsstore"]

    @pytest.fixture(params=destinations)
    def dest(self, request, tmpdir):
        if request.param == "hdf5":
            h5py = pytest.importorskip("h5py")
            fn = tmpdir.join("dest.h5")
            with h5py.File(str(fn), mode="w") as h5f:
                yield h5f
        elif request.param == "zarr":
            yield group(path="group2", zarr_version=3)
        elif request.param == "zarr_kvstore":
            store = KVStoreV3(dict())
            yield group(store, path="group2", zarr_version=3)
        elif request.param == "zarr_fsstore":
            fn = tmpdir.join("dest.zr3")
            store = FSStoreV3(str(fn), auto_mkdir=True)
            yield group(store, path="group2", zarr_version=3)
        elif request.param == "zarr_directorystore":
            fn = tmpdir.join("dest.zr3")
            store = DirectoryStoreV3(str(fn))
            yield group(store, path="group2", zarr_version=3)
        elif request.param == "zarr_sqlitestore":
            fn = tmpdir.join("dest.db")
            store = SQLiteStoreV3(str(fn))
            yield group(store, path="group2", zarr_version=3)

    def test_copy_array_create_options(self, source, dest):
        dest_h5py = dest.__module__.startswith("h5py.")

        # copy array, provide creation options
        compressor = Zlib(9)
        create_kws = dict(chunks=(10,))
        if dest_h5py:
            create_kws.update(
                compression="gzip", compression_opts=9, shuffle=True, fletcher32=True, fillvalue=42
            )
        else:
            # v3 case has no filters argument in zarr create_kws
            create_kws.update(compressor=compressor, fill_value=42, order="F")
        copy(source["foo/bar/baz"], dest, without_attrs=True, **create_kws)
        check_copied_array(
            source["foo/bar/baz"], dest["baz"], without_attrs=True, expect_props=create_kws
        )

    def test_copy_group_no_name(self, source, dest):
        if source.__module__.startswith("h5py"):
            with pytest.raises(TypeError):
                copy(source, dest)
        else:
            # For v3, dest.name will be inferred from source.name
            copy(source, dest)
            check_copied_group(source, dest[source.name.lstrip("/")])

        copy(source, dest, name="root")
        check_copied_group(source, dest["root"])



import warnings
from collections.abc import MutableMapping

from zarr._storage.store import Store, StoreV3
from zarr.util import json_dumps


class Attributes(MutableMapping):
    """Class providing access to user attributes on an array or group. Should not be
    instantiated directly, will be available via the `.attrs` property of an array or
    group.

    Parameters
    ----------
    store : MutableMapping
        The store in which to store the attributes.
    key : str, optional
        The key under which the attributes will be stored.
    read_only : bool, optional
        If True, attributes cannot be modified.
    cache : bool, optional
        If True (default), attributes will be cached locally.
    synchronizer : Synchronizer
        Only necessary if attributes may be modified from multiple threads or processes.

    """

    def __init__(
        self, store, key=".zattrs", read_only=False, cache=True, synchronizer=None, cached_dict=None
    ):
        self._version = getattr(store, "_store_version", 2)
        _Store = Store if self._version == 2 else StoreV3
        self.store = _Store._ensure_store(store)
        self.key = key
        self.read_only = read_only
        self.cache = cache
        self._cached_asdict = cached_dict if cache else None
        self.synchronizer = synchronizer

    def _get_nosync(self):
        try:
            data = self.store[self.key]
        except KeyError:
            d = dict()
            if self._version > 2:
                d["attributes"] = {}
        else:
            d = self.store._metadata_class.parse_metadata(data)
        return d

    def asdict(self):
        """Retrieve all attributes as a dictionary."""
        if self.cache and self._cached_asdict is not None:
            return self._cached_asdict
        d = self._get_nosync()
        if self._version == 3:
            d = d["attributes"]
        if self.cache:
            self._cached_asdict = d
        return d

    def refresh(self):
        """Refresh cached attributes from the store."""
        if self.cache:
            if self._version == 2:
                self._cached_asdict = self._get_nosync()
            else:
                self._cached_asdict = self._get_nosync()["attributes"]

    def __contains__(self, x):
        return x in self.asdict()

    def __getitem__(self, item):
        return self.asdict()[item]

    def _write_op(self, f, *args, **kwargs):
        # guard condition
        if self.read_only:
            raise PermissionError("attributes are read-only")

        # synchronization
        if self.synchronizer is None:
            return f(*args, **kwargs)
        else:
            with self.synchronizer[self.key]:
                return f(*args, **kwargs)

    def __setitem__(self, item, value):
        self._write_op(self._setitem_nosync, item, value)

    def _setitem_nosync(self, item, value):
        # load existing data
        d = self._get_nosync()

        # set key value
        if self._version == 2:
            d[item] = value
        else:
            d["attributes"][item] = value

        # _put modified data
        self._put_nosync(d)

    def __delitem__(self, item):
        self._write_op(self._delitem_nosync, item)

    def _delitem_nosync(self, key):
        # load existing data
        d = self._get_nosync()

        # delete key value
        if self._version == 2:
            del d[key]
        else:
            del d["attributes"][key]

        # _put modified data
        self._put_nosync(d)

    def put(self, d):
        """Overwrite all attributes with the key/value pairs in the provided dictionary
        `d` in a single operation."""
        if self._version == 2:
            self._write_op(self._put_nosync, d)
        else:
            self._write_op(self._put_nosync, dict(attributes=d))

    def _put_nosync(self, d):
        d_to_check = d if self._version == 2 else d["attributes"]
        if not all(isinstance(item, str) for item in d_to_check):
            # TODO: Raise an error for non-string keys
            # raise TypeError("attribute keys must be strings")
            warnings.warn(
                "only attribute keys of type 'string' will be allowed in the future",
                DeprecationWarning,
                stacklevel=2,
            )

            try:
                d_to_check = {str(k): v for k, v in d_to_check.items()}
            except TypeError as ex:  # pragma: no cover
                raise TypeError("attribute keys can not be stringified") from ex

            if self._version == 2:
                d = d_to_check
            else:
                d["attributes"] = d_to_check

        if self._version == 2:
            self.store[self.key] = json_dumps(d)
            if self.cache:
                self._cached_asdict = d
        else:
            try:
                meta_unparsed = self.store[self.key]
                # Cannot write the attributes directly to JSON, but have to
                # store it within the pre-existing attributes key of the v3
                # metadata.

                # Note: this changes the store.counter result in test_caching_on!

                meta = self.store._metadata_class.parse_metadata(meta_unparsed)
                if "attributes" in meta and "filters" in meta["attributes"]:
                    # need to preserve any existing "filters" attribute
                    d["attributes"]["filters"] = meta["attributes"]["filters"]
                meta["attributes"] = d["attributes"]
            except KeyError:
                meta = d
            self.store[self.key] = json_dumps(meta)
            if self.cache:
                self._cached_asdict = d["attributes"]

    # noinspection PyMethodOverriding
    def update(self, *args, **kwargs):
        """Update the values of several attributes in a single operation."""
        self._write_op(self._update_nosync, *args, **kwargs)

    def _update_nosync(self, *args, **kwargs):
        # load existing data
        d = self._get_nosync()

        # update
        if self._version == 2:
            d.update(*args, **kwargs)
        else:
            d["attributes"].update(*args, **kwargs)

        # _put modified data
        self._put_nosync(d)

    def keys(self):
        return self.asdict().keys()

    def __iter__(self):
        return iter(self.asdict())

    def __len__(self):
        return len(self.asdict())

    def _ipython_key_completions_(self):
        return sorted(self)



import os
import shutil
from collections import OrderedDict
from collections.abc import MutableMapping
from threading import Lock
from typing import Union, Dict, Any, Optional

from zarr.errors import (
    MetadataError,
    ReadOnlyError,
)
from zarr.util import buffer_size, json_loads, normalize_storage_path
from zarr.types import DIMENSION_SEPARATOR

from zarr._storage.absstore import ABSStoreV3  # noqa: F401
from zarr._storage.store import (  # noqa: F401
    _get_hierarchy_metadata,
    _get_metadata_suffix,
    _listdir_from_keys,
    _rename_from_keys,
    _rename_metadata_v3,
    _rmdir_from_keys,
    _rmdir_from_keys_v3,
    _path_to_prefix,
    _prefix_to_array_key,
    _prefix_to_group_key,
    array_meta_key,
    attrs_key,
    data_root,
    group_meta_key,
    meta_root,
    BaseStore,
    Store,
    StoreV3,
)
from zarr.storage import (
    DBMStore,
    ConsolidatedMetadataStore,
    DirectoryStore,
    FSStore,
    KVStore,
    LMDBStore,
    LRUStoreCache,
    MemoryStore,
    MongoDBStore,
    RedisStore,
    SQLiteStore,
    ZipStore,
    _getsize,
)

__doctest_requires__ = {
    ("RedisStore", "RedisStore.*"): ["redis"],
    ("MongoDBStore", "MongoDBStore.*"): ["pymongo"],
    ("LRUStoreCache", "LRUStoreCache.*"): ["s3fs"],
}


try:
    # noinspection PyUnresolvedReferences
    from zarr.codecs import Blosc

    default_compressor = Blosc()
except ImportError:  # pragma: no cover
    from zarr.codecs import Zlib

    default_compressor = Zlib()


Path = Union[str, bytes, None]
# allow MutableMapping for backwards compatibility
StoreLike = Union[BaseStore, MutableMapping]


class RmdirV3:
    """Mixin class that can be used to ensure override of any existing v2 rmdir class."""

    def rmdir(self, path: str = "") -> None:
        path = normalize_storage_path(path)
        _rmdir_from_keys_v3(self, path)  # type: ignore


class KVStoreV3(RmdirV3, KVStore, StoreV3):
    def list(self):
        return list(self._mutable_mapping.keys())

    def __setitem__(self, key, value):
        self._validate_key(key)
        super().__setitem__(key, value)

    def __eq__(self, other):
        return isinstance(other, KVStoreV3) and self._mutable_mapping == other._mutable_mapping


KVStoreV3.__doc__ = KVStore.__doc__


def _get_files_and_dirs_from_path(store, path):
    path = normalize_storage_path(path)

    files = []
    # add array metadata file if present
    array_key = _prefix_to_array_key(store, path)
    if array_key in store:
        files.append(os.path.join(store.path, array_key))

    # add group metadata file if present
    group_key = _prefix_to_group_key(store, path)
    if group_key in store:
        files.append(os.path.join(store.path, group_key))

    dirs = []
    # add array and group folders if present
    for d in [data_root + path, meta_root + path]:
        dir_path = os.path.join(store.path, d)
        if os.path.exists(dir_path):
            dirs.append(dir_path)
    return files, dirs


class FSStoreV3(FSStore, StoreV3):
    # FSStoreV3 doesn't use this (FSStore uses it within _normalize_key)
    _META_KEYS = ()

    def __setitem__(self, key, value):
        self._validate_key(key)
        super().__setitem__(key, value)

    def _default_key_separator(self):
        if self.key_separator is None:
            self.key_separator = "/"

    def list(self):
        return list(self.keys())

    def _normalize_key(self, key):
        key = normalize_storage_path(key).lstrip("/")
        return key.lower() if self.normalize_keys else key

    def getsize(self, path=None):
        size = 0
        if path is None or path == "":
            # size of both the data and meta subdirs
            dirs = []
            for d in ["data/root", "meta/root"]:
                dir_path = os.path.join(self.path, d)
                if os.path.exists(dir_path):
                    dirs.append(dir_path)
        elif path in self:
            # access individual element by full path
            return buffer_size(self[path])
        else:
            files, dirs = _get_files_and_dirs_from_path(self, path)
            for file in files:
                size += os.path.getsize(file)
        for d in dirs:
            size += self.fs.du(d, total=True, maxdepth=None)
        return size

    def setitems(self, values):
        if self.mode == "r":
            raise ReadOnlyError()
        values = {self._normalize_key(key): val for key, val in values.items()}

        # initialize the /data/root/... folder corresponding to the array!
        # Note: zarr.tests.test_core_v3.TestArrayWithFSStoreV3PartialRead fails
        # without this explicit creation of directories
        subdirectories = set(os.path.dirname(v) for v in values.keys())
        for subdirectory in subdirectories:
            data_dir = os.path.join(self.path, subdirectory)
            if not self.fs.exists(data_dir):
                self.fs.mkdir(data_dir)

        self.map.setitems(values)

    def rmdir(self, path=None):
        if self.mode == "r":
            raise ReadOnlyError()
        if path:
            for base in [meta_root, data_root]:
                store_path = self.dir_path(base + path)
                if self.fs.isdir(store_path):
                    self.fs.rm(store_path, recursive=True)

            # remove any associated metadata files
            sfx = _get_metadata_suffix(self)
            meta_dir = (meta_root + path).rstrip("/")
            array_meta_file = meta_dir + ".array" + sfx
            self.pop(array_meta_file, None)
            group_meta_file = meta_dir + ".group" + sfx
            self.pop(group_meta_file, None)
        else:
            store_path = self.dir_path(path)
            if self.fs.isdir(store_path):
                self.fs.rm(store_path, recursive=True)

    @property
    def supports_efficient_get_partial_values(self):
        return True

    def get_partial_values(self, key_ranges):
        """Get multiple partial values.
        key_ranges can be an iterable of key, range pairs,
        where a range specifies two integers range_start and range_length
        as a tuple, (range_start, range_length).
        range_length may be None to indicate to read until the end.
        range_start may be negative to start reading range_start bytes
        from the end of the file.
        A key may occur multiple times with different ranges.
        Inserts None for missing keys into the returned list."""
        results = []
        for key, (range_start, range_length) in key_ranges:
            key = self._normalize_key(key)
            path = self.dir_path(key)
            try:
                if range_start is None or range_length is None:
                    end = None
                else:
                    end = range_start + range_length
                result = self.fs.cat_file(path, start=range_start, end=end)
            except self.map.missing_exceptions:
                result = None
            results.append(result)
        return results


class MemoryStoreV3(MemoryStore, StoreV3):
    def __init__(
        self, root=None, cls=dict, dimension_separator: Optional[DIMENSION_SEPARATOR] = None
    ):
        if root is None:
            self.root = cls()
        else:
            self.root = root
        self.cls = cls
        self.write_mutex = Lock()
        self._dimension_separator = dimension_separator  # TODO: modify for v3?

    def __eq__(self, other):
        return (
            isinstance(other, MemoryStoreV3) and self.root == other.root and self.cls == other.cls
        )

    def __setitem__(self, key, value):
        self._validate_key(key)
        super().__setitem__(key, value)

    def list(self):
        return list(self.keys())

    def getsize(self, path: Path = None):
        return _getsize(self, path)

    def rename(self, src_path: Path, dst_path: Path):
        src_path = normalize_storage_path(src_path)
        dst_path = normalize_storage_path(dst_path)

        any_renamed = False
        for base in [meta_root, data_root]:
            if self.list_prefix(base + src_path):
                src_parent, src_key = self._get_parent(base + src_path)
                dst_parent, dst_key = self._require_parent(base + dst_path)

                if src_key in src_parent:
                    dst_parent[dst_key] = src_parent.pop(src_key)

                if base == meta_root:
                    # check for and move corresponding metadata
                    sfx = _get_metadata_suffix(self)
                    src_meta = src_key + ".array" + sfx
                    if src_meta in src_parent:
                        dst_meta = dst_key + ".array" + sfx
                        dst_parent[dst_meta] = src_parent.pop(src_meta)
                    src_meta = src_key + ".group" + sfx
                    if src_meta in src_parent:
                        dst_meta = dst_key + ".group" + sfx
                        dst_parent[dst_meta] = src_parent.pop(src_meta)
                any_renamed = True
        any_renamed = _rename_metadata_v3(self, src_path, dst_path) or any_renamed
        if not any_renamed:
            raise ValueError(f"no item {src_path} found to rename")

    def rmdir(self, path: Path = None):
        path = normalize_storage_path(path)
        if path:
            for base in [meta_root, data_root]:
                try:
                    parent, key = self._get_parent(base + path)
                    value = parent[key]
                except KeyError:
                    continue
                else:
                    if isinstance(value, self.cls):
                        del parent[key]

            # remove any associated metadata files
            sfx = _get_metadata_suffix(self)
            meta_dir = (meta_root + path).rstrip("/")
            array_meta_file = meta_dir + ".array" + sfx
            self.pop(array_meta_file, None)
            group_meta_file = meta_dir + ".group" + sfx
            self.pop(group_meta_file, None)
        else:
            # clear out root
            self.root = self.cls()


MemoryStoreV3.__doc__ = MemoryStore.__doc__


class DirectoryStoreV3(DirectoryStore, StoreV3):
    def list(self):
        return list(self.keys())

    def __eq__(self, other):
        return isinstance(other, DirectoryStoreV3) and self.path == other.path

    def __setitem__(self, key, value):
        self._validate_key(key)
        super().__setitem__(key, value)

    def getsize(self, path: Path = None):
        return _getsize(self, path)

    def rename(self, src_path, dst_path, metadata_key_suffix=".json"):
        store_src_path = normalize_storage_path(src_path)
        store_dst_path = normalize_storage_path(dst_path)

        dir_path = self.path
        any_existed = False
        for root_prefix in ["meta", "data"]:
            src_path = os.path.join(dir_path, root_prefix, "root", store_src_path)
            if os.path.exists(src_path):
                any_existed = True
                dst_path = os.path.join(dir_path, root_prefix, "root", store_dst_path)
                os.renames(src_path, dst_path)

        for suffix in [".array" + metadata_key_suffix, ".group" + metadata_key_suffix]:
            src_meta = os.path.join(dir_path, "meta", "root", store_src_path + suffix)
            if os.path.exists(src_meta):
                any_existed = True
                dst_meta = os.path.join(dir_path, "meta", "root", store_dst_path + suffix)
                dst_dir = os.path.dirname(dst_meta)
                if not os.path.exists(dst_dir):
                    os.makedirs(dst_dir)
                os.rename(src_meta, dst_meta)
        if not any_existed:
            raise FileNotFoundError("nothing found at src_path")

    def rmdir(self, path=None):
        store_path = normalize_storage_path(path)
        dir_path = self.path
        if store_path:
            for base in [meta_root, data_root]:
                dir_path = os.path.join(dir_path, base + store_path)
                if os.path.isdir(dir_path):
                    shutil.rmtree(dir_path)

            # remove any associated metadata files
            sfx = _get_metadata_suffix(self)
            meta_dir = (meta_root + path).rstrip("/")
            array_meta_file = meta_dir + ".array" + sfx
            self.pop(array_meta_file, None)
            group_meta_file = meta_dir + ".group" + sfx
            self.pop(group_meta_file, None)

        elif os.path.isdir(dir_path):
            shutil.rmtree(dir_path)


DirectoryStoreV3.__doc__ = DirectoryStore.__doc__


class ZipStoreV3(ZipStore, StoreV3):
    def list(self):
        return list(self.keys())

    def __eq__(self, other):
        return (
            isinstance(other, ZipStore)
            and self.path == other.path
            and self.compression == other.compression
            and self.allowZip64 == other.allowZip64
        )

    def __setitem__(self, key, value):
        self._validate_key(key)
        super().__setitem__(key, value)

    def getsize(self, path=None):
        path = normalize_storage_path(path)
        with self.mutex:
            children = self.list_prefix(data_root + path)
            children += self.list_prefix(meta_root + path)
            print(f"path={path}, children={children}")
            if children:
                size = 0
                for name in children:
                    info = self.zf.getinfo(name)
                    size += info.compress_size
                return size
            elif path in self:
                info = self.zf.getinfo(path)
                return info.compress_size
            else:
                return 0


ZipStoreV3.__doc__ = ZipStore.__doc__


class RedisStoreV3(RmdirV3, RedisStore, StoreV3):
    def list(self):
        return list(self.keys())

    def __setitem__(self, key, value):
        self._validate_key(key)
        super().__setitem__(key, value)


RedisStoreV3.__doc__ = RedisStore.__doc__


class MongoDBStoreV3(RmdirV3, MongoDBStore, StoreV3):
    def list(self):
        return list(self.keys())

    def __setitem__(self, key, value):
        self._validate_key(key)
        super().__setitem__(key, value)


MongoDBStoreV3.__doc__ = MongoDBStore.__doc__


class DBMStoreV3(RmdirV3, DBMStore, StoreV3):
    def list(self):
        return list(self.keys())

    def __setitem__(self, key, value):
        self._validate_key(key)
        super().__setitem__(key, value)


DBMStoreV3.__doc__ = DBMStore.__doc__


class LMDBStoreV3(RmdirV3, LMDBStore, StoreV3):
    def list(self):
        return list(self.keys())

    def __setitem__(self, key, value):
        self._validate_key(key)
        super().__setitem__(key, value)


LMDBStoreV3.__doc__ = LMDBStore.__doc__


class SQLiteStoreV3(SQLiteStore, StoreV3):
    def list(self):
        return list(self.keys())

    def getsize(self, path=None):
        # TODO: why does the query below not work in this case?
        #       For now fall back to the default _getsize implementation
        # size = 0
        # for _path in [data_root + path, meta_root + path]:
        #     c = self.cursor.execute(
        #         '''
        #         SELECT COALESCE(SUM(LENGTH(v)), 0) FROM zarr
        #         WHERE k LIKE (? || "%") AND
        #               0 == INSTR(LTRIM(SUBSTR(k, LENGTH(?) + 1), "/"), "/")
        #         ''',
        #         (_path, _path)
        #     )
        #     for item_size, in c:
        #         size += item_size
        # return size

        # fallback to default implementation for now
        return _getsize(self, path)

    def __setitem__(self, key, value):
        self._validate_key(key)
        super().__setitem__(key, value)

    def rmdir(self, path=None):
        path = normalize_storage_path(path)
        if path:
            for base in [meta_root, data_root]:
                with self.lock:
                    self.cursor.execute("DELETE FROM zarr WHERE k LIKE (? || '/%')", (base + path,))
            # remove any associated metadata files
            sfx = _get_metadata_suffix(self)
            meta_dir = (meta_root + path).rstrip("/")
            array_meta_file = meta_dir + ".array" + sfx
            self.pop(array_meta_file, None)
            group_meta_file = meta_dir + ".group" + sfx
            self.pop(group_meta_file, None)
        else:
            self.clear()


SQLiteStoreV3.__doc__ = SQLiteStore.__doc__


class LRUStoreCacheV3(RmdirV3, LRUStoreCache, StoreV3):
    def __init__(self, store, max_size: int):
        self._store = StoreV3._ensure_store(store)
        self._max_size = max_size
        self._current_size = 0
        self._keys_cache = None
        self._contains_cache = {}
        self._listdir_cache: Dict[Path, Any] = dict()
        self._values_cache: Dict[Path, Any] = OrderedDict()
        self._mutex = Lock()
        self.hits = self.misses = 0

    def list(self):
        return list(self.keys())

    def __setitem__(self, key, value):
        self._validate_key(key)
        super().__setitem__(key, value)


LRUStoreCacheV3.__doc__ = LRUStoreCache.__doc__


class ConsolidatedMetadataStoreV3(ConsolidatedMetadataStore, StoreV3):
    """A layer over other storage, where the metadata has been consolidated into
    a single key.

    The purpose of this class, is to be able to get all of the metadata for
    a given array in a single read operation from the underlying storage.
    See :func:`zarr.convenience.consolidate_metadata` for how to create this
    single metadata key.

    This class loads from the one key, and stores the data in a dict, so that
    accessing the keys no longer requires operations on the backend store.

    This class is read-only, and attempts to change the array metadata will
    fail, but changing the data is possible. If the backend storage is changed
    directly, then the metadata stored here could become obsolete, and
    :func:`zarr.convenience.consolidate_metadata` should be called again and the class
    re-invoked. The use case is for write once, read many times.

    .. note:: This is an experimental feature.

    Parameters
    ----------
    store: Store
        Containing the zarr array.
    metadata_key: str
        The target in the store where all of the metadata are stored. We
        assume JSON encoding.

    See Also
    --------
    zarr.convenience.consolidate_metadata, zarr.convenience.open_consolidated

    """

    def __init__(self, store: StoreLike, metadata_key=meta_root + "consolidated/.zmetadata"):
        self.store = StoreV3._ensure_store(store)

        # retrieve consolidated metadata
        meta = json_loads(self.store[metadata_key])

        # check format of consolidated metadata
        consolidated_format = meta.get("zarr_consolidated_format", None)
        if consolidated_format != 1:
            raise MetadataError(
                f"unsupported zarr consolidated metadata format: {consolidated_format}"
            )

        # decode metadata
        self.meta_store: Store = KVStoreV3(meta["metadata"])

    def rmdir(self, key):
        raise ReadOnlyError()


def _normalize_store_arg_v3(store: Any, storage_options=None, mode="r") -> BaseStore:
    # default to v2 store for backward compatibility
    zarr_version = getattr(store, "_store_version", 3)
    if zarr_version != 3:
        raise ValueError("store must be a version 3 store")
    if store is None:
        store = KVStoreV3(dict())
        # add default zarr.json metadata
        store["zarr.json"] = store._metadata_class.encode_hierarchy_metadata(None)
        return store
    if isinstance(store, os.PathLike):
        store = os.fspath(store)
    if FSStore._fsspec_installed():
        import fsspec

        if isinstance(store, fsspec.FSMap):
            return FSStoreV3(
                store.root,
                fs=store.fs,
                mode=mode,
                check=store.check,
                create=store.create,
                missing_exceptions=store.missing_exceptions,
                **(storage_options or {}),
            )
    if isinstance(store, str):
        if "://" in store or "::" in store:
            store = FSStoreV3(store, mode=mode, **(storage_options or {}))
        elif storage_options:
            raise ValueError("storage_options passed with non-fsspec path")
        elif store.endswith(".zip"):
            store = ZipStoreV3(store, mode=mode)
        elif store.endswith(".n5"):
            raise NotImplementedError("N5Store not yet implemented for V3")
            # return N5StoreV3(store)
        else:
            store = DirectoryStoreV3(store)
    else:
        store = StoreV3._ensure_store(store)

    if "zarr.json" not in store:
        # add default zarr.json metadata
        store["zarr.json"] = store._metadata_class.encode_hierarchy_metadata(None)
    return store
